"""Label subscription."""

from typing import Callable

from typeguard import typechecked

from kili.core.graphql.graphql_client import GraphQLClient
from kili.core.graphql.ws_graphql_client import SubscriptionGraphQLClient

from .subscriptions import GQL_LABEL_CREATED_OR_UPDATED


# pylint: disable=too-few-public-methods
class SubscriptionsLabel:
    """Set of Label subscriptions."""

    graphql_client: GraphQLClient

    @typechecked
    def label_created_or_updated(
        self, project_id: str, callback: Callable[[str, str], None]
    ) -> SubscriptionGraphQLClient:
        # pylint: disable=line-too-long
        """Subscribe a callback to a project, which is executed when a label is created or updated.

        Args:
            project_id: Identifier of the project
            callback: This function takes as input the id of the asset and its content.

        Returns:
            A subscription client.

        !!! example "Recipe"
            For more detailed examples on how to use Webhooks,
            See [the related recipe](https://github.com/kili-technology/kili-python-sdk/blob/main/recipes/webhooks_example.ipynb)
        """
        ws_endpoint = self.graphql_client.endpoint.replace("http", "ws")
        websocket = SubscriptionGraphQLClient(ws_endpoint)
        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        authorization = f"X-API-Key: {self.api_key}"  # type: ignore  # pylint: disable=no-member
        headers["Authorization"] = authorization
        variables = {"projectID": project_id}
        websocket.subscribe(
            GQL_LABEL_CREATED_OR_UPDATED,
            variables=variables,
            callback=callback,
            headers=headers,
            authorization=authorization,
        )
        return websocket



"""Service for exporting kili objects."""

from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Type, Union

from typing_extensions import get_args

from kili.domain.asset import AssetId
from kili.domain.project import ProjectId
from kili.services.export.format.base import AbstractExporter, ExportParams
from kili.services.export.format.coco import CocoExporter
from kili.services.export.format.geojson import GeoJsonExporter
from kili.services.export.format.kili import KiliExporter
from kili.services.export.format.llm import LLMExporter
from kili.services.export.format.voc import VocExporter
from kili.services.export.format.yolo import YoloExporter
from kili.services.export.logger import get_logger
from kili.services.export.repository import SDKContentRepository
from kili.services.export.types import CocoAnnotationModifier, ExportType, LabelFormat, SplitOption
from kili.services.types import LogLevel

if TYPE_CHECKING:
    from kili.client import Kili


def export_labels(  # pylint: disable=too-many-arguments, too-many-locals
    kili: "Kili",
    asset_ids: Optional[List[AssetId]],
    project_id: ProjectId,
    export_type: ExportType,
    label_format: LabelFormat,
    split_option: SplitOption,
    single_file: bool,
    output_file: Optional[str],
    disable_tqdm: Optional[bool],
    log_level: LogLevel,
    with_assets: bool,
    annotation_modifier: Optional[CocoAnnotationModifier],
    asset_filter_kwargs: Optional[Dict[str, object]],
    normalized_coordinates: Optional[bool],
    label_type_in: Optional[List[str]],
    include_sent_back_labels: Optional[bool],
) -> Optional[List[Dict[str, Union[List[str], str]]]]:
    """Export the selected assets into the required format, and save it into a file archive."""
    kili.kili_api_gateway.get_project(project_id, ["id"])

    include_sent_back_labels = (
        include_sent_back_labels
        if include_sent_back_labels is not None
        else (label_format != "llm_v1")
    )

    export_params = ExportParams(
        assets_ids=asset_ids,
        project_id=project_id,
        export_type=export_type,
        label_format=label_format,
        split_option=split_option,
        single_file=single_file,
        output_file=Path(output_file) if output_file is not None else None,
        with_assets=with_assets,
        annotation_modifier=annotation_modifier,
        asset_filter_kwargs=asset_filter_kwargs,
        normalized_coordinates=normalized_coordinates,
        label_type_in=label_type_in,
        include_sent_back_labels=include_sent_back_labels,
    )

    logger = get_logger(log_level)

    content_repository = SDKContentRepository(kili.api_endpoint, http_client=kili.http_client)

    if label_format in get_args(LabelFormat):
        format_exporter_selector_mapping: Dict[str, Type[AbstractExporter]] = {
            "raw": KiliExporter,
            "kili": KiliExporter,
            "coco": CocoExporter,
            "yolo_v4": YoloExporter,
            "yolo_v5": YoloExporter,
            "yolo_v7": YoloExporter,
            "yolo_v8": YoloExporter,
            "pascal_voc": VocExporter,
            "geojson": GeoJsonExporter,
            "llm_v1": LLMExporter,
            "llm_dynamic_v1": LLMExporter,
        }
        assert set(format_exporter_selector_mapping.keys()) == set(
            get_args(LabelFormat)
        )  # ensures full mapping
        exporter_class = format_exporter_selector_mapping[label_format]
        return exporter_class(
            export_params, kili, logger, disable_tqdm, content_repository
        ).export_project()

    raise ValueError(f'Label format "{label_format}" is not implemented or does not exist.')



"""Set of common functions used by different export formats."""

import warnings
from typing import Dict, List, Optional

from kili.adapters.http_client import HttpClient
from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.constants import QUERY_BATCH_SIZE
from kili.core.helpers import (
    get_response_json,
    log_raise_for_status,
    validate_category_search_query,
)
from kili.core.utils.pagination import batcher
from kili.domain.asset import AssetFilters, AssetId
from kili.domain.project import ProjectId
from kili.services.export.types import ExportType
from kili.use_cases.asset.media_downloader import get_download_assets_function

COMMON_FIELDS = [
    "id",
    "externalId",
    "content",
    "jsonContent",
    "jsonMetadata",
    "status",
    "pageResolutions.pageNumber",
    "pageResolutions.height",
    "pageResolutions.width",
    "pageResolutions.rotation",
    "resolution.height",
    "resolution.width",
]

DEFAULT_FIELDS = [
    *COMMON_FIELDS,
    "labels.jsonResponse",
    "labels.author.id",
    "labels.author.email",
    "labels.author.firstname",
    "labels.author.lastname",
    "labels.createdAt",
    "labels.isLatestLabelForUser",
    "labels.isSentBackToQueue",
    "labels.labelType",
    "labels.modelName",
]
LATEST_LABEL_FIELDS = [
    *COMMON_FIELDS,
    "latestLabel.jsonResponse",
    "latestLabel.author.id",
    "latestLabel.author.email",
    "latestLabel.author.firstname",
    "latestLabel.author.lastname",
    "latestLabel.createdAt",
    "latestLabel.isLatestLabelForUser",
    "latestLabel.isSentBackToQueue",
    "latestLabel.labelType",
    "latestLabel.modelName",
]


def attach_name_to_assets_labels_author(assets: List[Dict], export_type: ExportType):
    """Adds `name` field for author, by concatenating his/her first and last name."""
    for asset in assets:
        if export_type == "latest":
            latest_label = asset["latestLabel"]
            if latest_label:
                firstname = latest_label["author"]["firstname"]
                lastname = latest_label["author"]["lastname"]
                latest_label["author"]["name"] = f"{firstname} {lastname}"
            continue
        for label in asset.get("labels", []):
            firstname = label["author"]["firstname"]
            lastname = label["author"]["lastname"]
            label["author"]["name"] = f"{firstname} {lastname}"


THRESHOLD_WARN_MANY_ASSETS = 1000


# pylint: disable=too-many-arguments, too-many-locals, missing-type-doc
def fetch_assets(
    kili,
    project_id: str,
    asset_ids: Optional[List[AssetId]],
    export_type: ExportType,
    label_type_in: Optional[List[str]],
    disable_tqdm: Optional[bool],
    download_media: bool,
    local_media_dir: Optional[str],
    asset_filter_kwargs: Optional[Dict[str, object]],
) -> List[Dict]:
    """Fetches assets.

    Fetches assets where ID are in asset_ids if the list has more than one element,
    else all the assets of the project. If download media is passed, the media are
    downloaded into the `$HOME/.cache` folder.

    Args:
        kili: Kili instance
        project_id: project id
        asset_ids: list of asset IDs
        export_type: type of export (latest label or all labels)
        label_type_in: types of label to fetch (default, reviewed, ...)
        disable_tqdm: tell to disable tqdm
        download_media: tell to download the media in the cache folder.
        local_media_dir: Directory where the media are downloaded if `download_media` is True.
        asset_filter_kwargs: Optional dictionary of arguments to filter the assets to export.

    Returns:
        List of fetched assets.
    """
    fields = get_fields_to_fetch(export_type)
    asset_filter_kwargs = asset_filter_kwargs or {}
    asset_where_params = {
        "project_id": project_id,
        "label_type_in": label_type_in,
        "consensus_mark_gte": asset_filter_kwargs.pop("consensus_mark_gte", None),
        "consensus_mark_lte": asset_filter_kwargs.pop("consensus_mark_lte", None),
        "external_id_strictly_in": asset_filter_kwargs.pop("external_id_strictly_in", None)
        or asset_filter_kwargs.pop("external_id_contains", None),
        "external_id_in": asset_filter_kwargs.pop("external_id_in", None),
        "honeypot_mark_gte": asset_filter_kwargs.pop("honeypot_mark_gte", None),
        "honeypot_mark_lte": asset_filter_kwargs.pop("honeypot_mark_lte", None),
        "label_author_in": asset_filter_kwargs.pop("label_author_in", None),
        "label_labeler_in": asset_filter_kwargs.pop("label_labeler_in", None),
        "label_labeler_not_in": asset_filter_kwargs.pop("label_labeler_not_in", None),
        "label_reviewer_in": asset_filter_kwargs.pop("label_reviewer_in", None),
        "label_reviewer_not_in": asset_filter_kwargs.pop("label_reviewer_not_in", None),
        "assignee_in": asset_filter_kwargs.pop("assignee_in", None),
        "assignee_not_in": asset_filter_kwargs.pop("assignee_not_in", None),
        "skipped": asset_filter_kwargs.pop("skipped", None),
        "status_in": asset_filter_kwargs.pop("status_in", None),
        "label_category_search": asset_filter_kwargs.pop("label_category_search", None),
        "created_at_gte": asset_filter_kwargs.pop("created_at_gte", None),
        "created_at_lte": asset_filter_kwargs.pop("created_at_lte", None),
        "issue_type": asset_filter_kwargs.pop("issue_type", None),
        "issue_status": asset_filter_kwargs.pop("issue_status", None),
        "inference_mark_gte": asset_filter_kwargs.pop("inference_mark_gte", None),
        "inference_mark_lte": asset_filter_kwargs.pop("inference_mark_lte", None),
        "metadata_where": asset_filter_kwargs.pop("metadata_where", None),
    }

    if asset_filter_kwargs:
        raise NameError(f"Unknown asset filter arguments: {list(asset_filter_kwargs.keys())}")

    if asset_where_params.get("label_category_search"):
        validate_category_search_query(asset_where_params["label_category_search"])  # type: ignore

    if asset_ids is not None and len(asset_ids) > 0:
        asset_where_params["asset_id_in"] = asset_ids

    filters = AssetFilters(**asset_where_params)

    if download_media:
        count = kili.kili_api_gateway.count_assets(filters)
        if count > THRESHOLD_WARN_MANY_ASSETS:
            warnings.warn(
                f"Downloading many assets ({count}). This might take a while. Consider"
                " disabling assets download in the options.",
                stacklevel=3,
            )

    options = QueryOptions(disable_tqdm=disable_tqdm)
    download_media_function, fields = get_download_assets_function(
        kili.kili_api_gateway, download_media, fields, ProjectId(project_id), local_media_dir
    )
    assets_gen = kili.kili_api_gateway.list_assets(filters, fields, options)

    if (label_type_in is not None) and (len(label_type_in) > 0):
        if export_type == "latest":
            assets_gen = filter(lambda asset: asset.get("latestLabel") is not None, assets_gen)
            assets_gen = filter(
                lambda asset: asset["latestLabel"].get("labelType") in label_type_in, assets_gen
            )
        else:
            assets_gen = filter(lambda asset: asset.get("labels") is not None, assets_gen)
            assets_gen = filter(
                lambda asset: any(
                    label.get("labelType") in label_type_in for label in asset["labels"]
                ),
                assets_gen,
            )
            assets_gen = (
                {
                    **asset,
                    "labels": [
                        label
                        for label in asset["labels"]
                        if label.get("labelType") in label_type_in
                    ],
                }
                for asset in assets_gen
            )

    if download_media_function is not None:
        assets: List[Dict] = []
        # TODO: modify download_media function so it can take a generator of assets
        for assets_batch in batcher(assets_gen, QUERY_BATCH_SIZE):
            assets.extend(download_media_function(assets_batch))
    else:
        assets = list(assets_gen)
    attach_name_to_assets_labels_author(assets, export_type)
    return assets


def get_fields_to_fetch(export_type: ExportType):
    """Return the fields to fetch depending on the export type."""
    if export_type == "latest":
        return LATEST_LABEL_FIELDS
    return DEFAULT_FIELDS


def is_geotiff_asset_with_lat_lon_coords(asset: Dict, http_client: HttpClient) -> bool:
    """Check if asset is a geotiff with lat/lon coordinates."""
    if "jsonContent" not in asset:
        return False

    if isinstance(asset["jsonContent"], str) and asset["jsonContent"].startswith("http"):
        response = http_client.get(asset["jsonContent"], timeout=30)
        log_raise_for_status(response)
        json_content = get_response_json(response)

    else:
        json_content = asset["jsonContent"]

    return (
        isinstance(json_content, List)
        and len(json_content) > 0
        and isinstance(json_content[0], Dict)
        and json_content[0].get("useClassicCoordinates") is False
        and "epsg" in json_content[0]
        and json_content[0]["epsg"] != "TiledImage"
    )



"""Client presentation methods for questions."""

from itertools import repeat
from typing import Dict, List, Literal, Optional

from typeguard import typechecked

from kili.domain.asset import AssetExternalId, AssetId
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.use_cases.question import QuestionToCreateUseCaseInput, QuestionUseCases

from .base import BaseClientMethods


class QuestionClientMethods(BaseClientMethods):
    """Client presentation methods for questions."""

    @typechecked
    def create_questions(
        self,
        project_id: str,
        text_array: ListOrTuple[Optional[str]],
        asset_id_array: Optional[ListOrTuple[str]] = None,
        asset_external_id_array: Optional[ListOrTuple[str]] = None,
    ) -> List[Dict[Literal["id"], str]]:
        # pylint:disable=line-too-long
        """Create questions.

        Args:
            project_id: Id of the project.
            text_array: List of question strings.
            asset_id_array: List of the assets to add the questions to.
            asset_external_id_array: List of the assets to add the questions to. Used if `asset_id_array` is not given.

        Returns:
            A list of dictionaries with the `id` key of the created questions.
        """
        use_case_questions = [
            QuestionToCreateUseCaseInput(
                text=text,
                asset_id=AssetId(asset_id) if asset_id else None,
                asset_external_id=AssetExternalId(asset_external_id) if asset_external_id else None,
            )
            for (text, asset_id, asset_external_id) in zip(
                text_array, asset_id_array or repeat(None), asset_external_id_array or repeat(None)
            )
        ]
        question_ids = QuestionUseCases(self.kili_api_gateway).create_questions(
            project_id=ProjectId(project_id),
            questions=use_case_questions,
        )

        return [{"id": question_id} for question_id in question_ids]



"""Client presentation methods for labels."""

# pylint: disable=too-many-lines
import warnings
from typing import (
    Dict,
    List,
    Optional,
    Union,
    cast,
)

from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.domain.asset import AssetExternalId, AssetFilters, AssetId, AssetStatus
from kili.domain.label import LabelType
from kili.domain.llm import (
    AzureOpenAICredentials,
    ChatItemDict,
    ModelDict,
    ModelToCreateInput,
    ModelToUpdateInput,
    ModelType,
    OpenAISDKCredentials,
    OrganizationModelFilters,
    ProjectModelDict,
    ProjectModelFilters,
    ProjectModelToCreateInput,
    ProjectModelToUpdateInput,
)
from kili.domain.project import ProjectId
from kili.llm.services.export import export
from kili.services.export.exceptions import NoCompatibleJobError
from kili.use_cases.asset.utils import AssetUseCasesUtils
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class LlmClientMethods:
    def __init__(self, kili_api_gateway: KiliAPIGateway):
        self.kili_api_gateway = kili_api_gateway

    def export(
        self,
        project_id: str,
        disable_tqdm: Optional[bool] = False,
        asset_ids: Optional[List[str]] = None,
        external_ids: Optional[List[str]] = None,
        include_sent_back_labels: Optional[bool] = False,
        label_type_in: Optional[List[LabelType]] = None,
        status_in: Optional[List[AssetStatus]] = None,
    ) -> Optional[List[Dict[str, Union[List[str], str]]]]:
        # pylint: disable=line-too-long
        """Returns an export of llm assets with valid labels.

        Args:
            project_id: Identifier of the project.
            asset_ids: Optional list of the assets internal IDs from which to export the labels.
            disable_tqdm: Disable the progress bar if True.
            external_ids: Optional list of the assets external IDs from which to export the labels.
            include_sent_back_labels: Include sent back labels if True.
            label_type_in: Optional types of label to fetch, by default ["DEFAULT", "REVIEW"].
            status_in: Returned assets should have a status that belongs to that list, if given.
                Possible choices: `TODO`, `ONGOING`, `LABELED`, `TO_REVIEW` or `REVIEWED`.
        !!! Example
            ```python
            kili.llm.export("your_project_id")
            ```
        """
        if asset_ids and external_ids:
            raise ValueError("You cannot provide both asset_ids and external_ids")

        if external_ids is not None and asset_ids is None:
            id_map = AssetUseCasesUtils(self.kili_api_gateway).infer_ids_from_external_ids(
                asset_external_ids=cast(List[AssetExternalId], external_ids),
                project_id=ProjectId(project_id),
            )
            resolved_asset_ids = [id_map[AssetExternalId(i)] for i in external_ids]
        else:
            resolved_asset_ids = (
                list(AssetId(asset_id) for asset_id in asset_ids) if asset_ids else None
            )

        label_type_in = label_type_in or ["DEFAULT", "REVIEW"]

        asset_filter = AssetFilters(
            project_id=ProjectId(project_id),
            asset_id_in=resolved_asset_ids,
            label_type_in=label_type_in,
            status_in=status_in,
        )

        try:
            return export(
                self.kili_api_gateway,
                project_id=ProjectId(project_id),
                asset_filter=asset_filter,
                disable_tqdm=disable_tqdm,
                include_sent_back_labels=include_sent_back_labels,
                label_type_in=label_type_in,
            )
        except NoCompatibleJobError as excp:
            warnings.warn(str(excp), stacklevel=2)
            return None

    def create_model(self, organization_id: str, model: dict) -> ModelDict:
        # pylint: disable=line-too-long
        """Create a new model in an organization.

        Args:
            organization_id: Identifier of the organization.
            model: A dictionary representing the model to create, containing:
                - `name`: Name of the model.
                - `type`: Type of the model, one of:
                    - `"AZURE_OPEN_AI"`
                    - `"OPEN_AI_SDK"`
                - `credentials`: Credentials required for the model, depending on the type:
                    - For `"AZURE_OPEN_AI"` type:
                        - `api_key`: The API key for Azure OpenAI.
                        - `deployment_id`: The deployment ID within Azure.
                        - `endpoint`: The endpoint URL of the Azure OpenAI service.
                    - For `"OPEN_AI_SDK"` type:
                        - `api_key`: The API key for OpenAI SDK.
                        - `endpoint`: The endpoint URL of the OpenAI SDK service.

        Returns:
            A dictionary containing the created model's details.

        Examples:
            >>> # Example of creating an OpenAI SDK model
            >>> model_data = {
            ...     "name": "My OpenAI SDK Model",
            ...     "type": "OPEN_AI_SDK",
            ...     "credentials": {
            ...         "api_key": "your_open_ai_api_key",
            ...         "endpoint": "your_open_ai_endpoint"
            ...     }
            ... }
            >>> kili.llm.create_model(organization_id="your_organization_id", model=model_data)
        """
        credentials_data = model["credentials"]
        model_type = ModelType(model["type"])

        if model_type == ModelType.AZURE_OPEN_AI:
            credentials = AzureOpenAICredentials(**credentials_data)
        elif model_type == ModelType.OPEN_AI_SDK:
            credentials = OpenAISDKCredentials(**credentials_data)
        else:
            raise ValueError(f"Unsupported model type: {model['type']}")

        model_input = ModelToCreateInput(
            credentials=credentials,
            name=model["name"],
            type=model_type,
            organization_id=organization_id,
        )
        return self.kili_api_gateway.create_model(model=model_input)

    def models(self, organization_id: str, fields: Optional[List[str]] = None) -> List[ModelDict]:
        # pylint: disable=line-too-long
        """List models in an organization.

        Args:
            organization_id: Identifier of the organization.
            fields: All the fields to request among the possible fields for the models.
                Defaults to ["id", "credentials", "name", "type"].

        Returns:
            A list of models.

        Examples:
            >>> kili.llm.models(organization_id="your_organization_id")
        """
        converted_filters = OrganizationModelFilters(
            organization_id=organization_id,
        )

        return list(self.kili_api_gateway.list_models(filters=converted_filters, fields=fields))

    def model(self, model_id: str, fields: Optional[List[str]] = None) -> ModelDict:
        # pylint: disable=line-too-long
        """Retrieve a specific model.

        Args:
            model_id: Identifier of the model.
            fields: All the fields to request among the possible fields for the models.
                Defaults to ["id", "credentials", "name", "type"].

        Returns:
            A dictionary representing the model.

        Examples:
            >>> kili.llm.model(model_id="your_model_id")
        """
        return self.kili_api_gateway.get_model(
            model_id=model_id,
            fields=fields,
        )

    def update_properties_in_model(self, model_id: str, model: dict) -> ModelDict:
        # pylint: disable=line-too-long
        """Update properties of an existing model.

        Args:
            model_id: Identifier of the model to update.
            model: A dictionary containing the properties to update, which may include:
                - `name`: New name of the model.
                - `credentials`: Updated credentials for the model, depending on the type:
                    - For `"AZURE_OPEN_AI"` type:
                        - `api_key`: The API key for Azure OpenAI.
                        - `deployment_id`: The deployment ID within Azure.
                        - `endpoint`: The endpoint URL of the Azure OpenAI service.
                    - For `"OPEN_AI_SDK"` type:
                        - `api_key`: The API key for OpenAI SDK.
                        - `endpoint`: The endpoint URL of the OpenAI SDK service.

        Returns:
            A dictionary containing the updated model's details.

        Examples:
            >>> # Update the name of a model
            >>> kili.llm.update_properties_in_model(
            ...     model_id="your_model_id",
            ...     model={"name": "Updated Model Name"}
            ... )
        """
        credentials_data = model.get("credentials")
        credentials = None

        if credentials_data:
            existing_model = self.kili_api_gateway.get_model(
                model_id=model_id, fields=["id", "type"]
            )
            if not existing_model:
                raise ValueError(f"Model with id {model_id} not found")
            model_type = ModelType(existing_model["type"])

            if model_type == ModelType.AZURE_OPEN_AI:
                credentials = AzureOpenAICredentials(**credentials_data)
            elif model_type == ModelType.OPEN_AI_SDK:
                credentials = OpenAISDKCredentials(**credentials_data)
            else:
                raise ValueError(f"Unsupported model type: {model_type}")

        model_input = ModelToUpdateInput(
            credentials=credentials,
            name=model.get("name"),
        )
        return self.kili_api_gateway.update_properties_in_model(
            model_id=model_id, model=model_input
        )

    def delete_model(self, model_id: str) -> bool:
        # pylint: disable=line-too-long
        """Delete a model from an organization.

        Args:
            model_id: Identifier of the model to delete.

        Returns:
            A dictionary indicating the result of the deletion.

        Examples:
            >>> kili.llm.delete_model(model_id="your_model_id")
        """
        return self.kili_api_gateway.delete_model(model_id=model_id)

    def create_project_model(
        self, project_id: str, model_id: str, configuration: dict
    ) -> ProjectModelDict:
        # pylint: disable=line-too-long
        """Associate a model with a project.

        Args:
            project_id: Identifier of the project.
            model_id: Identifier of the model to associate.
            configuration: Configuration parameters for the project model.

        Returns:
            A dictionary containing the created project model's details.

        Examples:
            >>> configuration = {
            ...     # Configuration details specific to your use case
            ... }
            >>> kili.llm.create_project_model(
            ...     project_id="your_project_id",
            ...     model_id="your_model_id",
            ...     configuration={"temperature": 0.7}
            ... )
        """
        project_model_input = ProjectModelToCreateInput(
            project_id=project_id, model_id=model_id, configuration=configuration
        )
        return self.kili_api_gateway.create_project_model(project_model=project_model_input)

    def project_models(
        self, project_id: str, filters: Optional[Dict] = None, fields: Optional[List[str]] = None
    ) -> List[ProjectModelDict]:
        """List models associated with a project.

        Args:
            project_id: Identifier of the project.
            filters: Optional filters to apply. Possible keys:
                - `model_id`: Identifier of a specific model to filter by.
            fields: All the fields to request among the possible fields for the project models.
                Defaults to ["configuration", "id", "model.credentials", "model.name", "model.type"].

        Returns:
            A list of project models.

        Examples:
            >>> kili.llm.project_models(project_id="your_project_id")
        """
        converted_filters = ProjectModelFilters(
            project_id=project_id,
            model_id=filters["model_id"] if filters and "model_id" in filters else None,
        )

        return list(
            self.kili_api_gateway.list_project_models(
                filters=converted_filters,
                fields=fields,
            )
        )

    def update_project_model(self, project_model_id: str, configuration: dict) -> ProjectModelDict:
        """Update the configuration of a project model.

        Args:
            project_model_id: Identifier of the project model to update.
            configuration: New configuration parameters.

        Returns:
            A dictionary containing the updated project model's details.

        Examples:
            >>> configuration = {
            ...     # Updated configuration details
            ... }
            >>> kili.llm.update_project_model(
            ...     project_model_id="your_project_model_id",
            ...     configuration=configuration
            ... )
        """
        project_model_input = ProjectModelToUpdateInput(configuration=configuration)
        return self.kili_api_gateway.update_project_model(
            project_model_id=project_model_id, project_model=project_model_input
        )

    def delete_project_model(self, project_model_id: str) -> bool:
        """Delete a project model.

        Args:
            project_model_id: Identifier of the project model to delete.

        Returns:
            A dictionary indicating the result of the deletion.

        Examples:
            >>> kili.llm.delete_project_model(project_model_id="your_project_model_id")
        """
        return self.kili_api_gateway.delete_project_model(project_model_id)

    def create_conversation(self, project_id: str, prompt: str) -> List[ChatItemDict]:
        # pylint: disable=line-too-long
        """Create a new conversation in an LLM project starting with a user's prompt.

        This method initiates a new conversation in the specified project by:
        - Creating an LLM asset and label associated with the current user.
        - Adding the user's prompt as the first chat item.
        - Automatically generating assistant responses using the project's models.

        Args:
            project_id: The identifier of the project where the conversation will be created.
            prompt: The initial prompt or message from the user to start the conversation.

        Returns:
            A list of chat items in the conversation, including the user's prompt and the assistant's responses.

        Examples:
            >>> PROMPT = "Hello, how can I improve my coding skills?"
            >>> chat_items = kili.llm.create_conversation(project_id="your_project_id", prompt=PROMPT)

        Notes:
            - The first chat item corresponds to the user's prompt.
            - The subsequent chat items are assistant responses generated by the project's models.
            - An LLM asset and a label are created in the project with labelType "PREDICTION".
        """
        user_id = self.kili_api_gateway.get_current_user(["id"])["id"]
        llm_asset = self.kili_api_gateway.create_llm_asset(
            project_id=project_id,
            author_id=user_id,
            label_type="PREDICTION",
        )
        asset_id = llm_asset["id"]
        label_id = llm_asset["latestLabel"]["id"]
        return self.kili_api_gateway.create_chat_item(
            asset_id=asset_id, label_id=label_id, prompt=prompt
        )



"""Common code for the GeoJson exporter."""

import json
from pathlib import Path
from typing import Dict, List

from kili.domain.ontology import JobMLTask, JobTool
from kili.services.export.exceptions import NotCompatibleInputType, NotCompatibleOptions
from kili.services.export.format.base import AbstractExporter
from kili.services.export.tools import is_geotiff_asset_with_lat_lon_coords
from kili.services.types import Job
from kili.utils.labels.geojson import kili_json_response_to_feature_collection
from kili.utils.tqdm import tqdm


class GeoJsonExporter(AbstractExporter):
    """Common code for GeoJson exporter."""

    def _check_arguments_compatibility(self) -> None:
        """Checks if the export label format is compatible with the export options."""
        if self.normalized_coordinates is not None:
            raise NotCompatibleOptions(
                "The GeoJson annotation format can only be exported with lat/long coordinates."
                " Please set `normalized_coordinates` to None."
            )

        if self.split_option != "merged":
            raise NotCompatibleOptions(
                "The current implementation only supports merged annotations."
            )

        if self.single_file:
            raise NotCompatibleOptions(
                "The GeoJson annotation format cannot be exported into a single file."
            )

    def _check_project_compatibility(self) -> None:
        """Checks if the export label format is compatible with the project."""
        if self.project["inputType"] != "IMAGE":
            raise NotCompatibleInputType(
                f"Project with input type '{self.project['inputType']}' not compatible with"
                " GeoJson export format."
            )

    def _is_job_compatible(self, job: Job) -> bool:
        """Check if the export label format is compatible with the job."""
        if "tools" not in job:
            return False

        if job["mlTask"] != JobMLTask.OBJECT_DETECTION:
            return False

        compatible_tools = {
            JobTool.RECTANGLE,
            JobTool.POLYGON,
            JobTool.SEMANTIC,
            JobTool.MARKER,
            JobTool.POLYLINE,
        }

        return all(
            tool in compatible_tools
            for tool in job["tools"]  # pyright: ignore[reportGeneralTypeIssues]
        )

    def process_and_save(self, assets: List[Dict], output_filename: Path) -> None:
        self.logger.info("Exporting to GeoJson format")

        labels_folder = self.base_folder / "labels"
        labels_folder.mkdir(parents=True, exist_ok=True)

        geotiff_assets = [
            asset
            for asset in assets
            if is_geotiff_asset_with_lat_lon_coords(asset, self.kili.http_client)
        ]
        if len(geotiff_assets) < len(assets):
            self.logger.warning(
                f"Among {len(assets)} assets, only {len(geotiff_assets)} are geotiff assets and"
                " will be exported."
            )

        for asset in tqdm(geotiff_assets, disable=self.disable_tqdm):
            _process_asset(asset, labels_folder)

        self.create_readme_kili_file(self.export_root_folder)
        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)


def _process_asset(asset: Dict, labels_folder: Path) -> None:
    geojson_feature_collection = kili_json_response_to_feature_collection(
        asset["latestLabel"]["jsonResponse"]
    )
    filepath = labels_folder / f'{asset["externalId"]}.geojson'
    filepath.parent.mkdir(parents=True, exist_ok=True)
    with open(filepath, "w", encoding="utf-8") as file:
        json.dump(geojson_feature_collection, file)



"""Set of common functions used by plugins."""

from pathlib import Path

DUPLICATE_ERROR_SUBSTRING = "an entity Plugin already exists with value"
DUPLICATE_ACTIVATE_ERROR_SUBSTRING = "an entity Project already exists with value"


def check_errors_plugin_upload(result, plugin_path: Path, plugin_name: str):
    """check_errors_plugin_upload.

    Check if the error during upload is because the plugin already
    exists and print the upload function for the user.
    """
    if "errors" in result and DUPLICATE_ERROR_SUBSTRING in result["errors"][0]["message"]:
        print(
            "Hint: A plugin with this name already exists, "
            "if you want to override it you can use the command "
            f'kili.update_plugin(plugin_path="{plugin_path}", plugin_name="{plugin_name}")'
        )


def check_errors_plugin_activation(result, plugin_name: str, project_id: str):
    """check_errors_plugin_activation.

    Check if the error during upload is because the plugin already
    exists and print the upload function for the user.
    """
    has_errors = "errors" in result
    if has_errors and DUPLICATE_ACTIVATE_ERROR_SUBSTRING in result["errors"][0]["message"]:
        print(
            f'Warning: A plugin with the name "{plugin_name}" is',
            f'already activated on the project with id "{project_id}"',
        )
        return has_errors, True

    return has_errors, False



"""Copy project implementation."""

import itertools
import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.constants import QUERY_BATCH_SIZE
from kili.core.utils.pagination import batcher
from kili.domain.asset import AssetFilters
from kili.domain.label import LabelFilters
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.use_cases.asset.media_downloader import get_download_assets_function
from kili.utils.tempfile import TemporaryDirectory
from kili.utils.tqdm import tqdm

if TYPE_CHECKING:
    from kili.client import Kili


class ProjectCopier:  # pylint: disable=too-few-public-methods
    """Class for copying an existing project."""

    FIELDS_PROJECT = (
        "title",
        "inputType",
        "description",
        "id",
        "dataConnections.id",
    )
    FIELDS_JSON_INTERFACE = ("jsonInterface",)
    FIELDS_QUALITY_SETTINGS = (
        "canSkipAsset",
        "consensusTotCoverage",
        "minConsensusSize",
        "reviewCoverage",
        "secondsToLabelBeforeAutoAssign",
        "useHoneyPot",
    )

    def __init__(self, kili: "Kili") -> None:
        self.disable_tqdm = False
        self.kili = kili

    def copy_project(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        from_project_id: str,
        title: Optional[str],
        description: Optional[str],
        copy_json_interface: bool,
        copy_quality_settings: bool,
        copy_members: bool,
        copy_assets: bool,
        copy_labels: bool,
        disable_tqdm: Optional[bool],
    ) -> str:
        """Copy an existing project."""
        self.disable_tqdm = disable_tqdm

        logging.basicConfig()
        logger = logging.getLogger("kili.services.copy_project")
        logger.setLevel(logging.INFO)

        if not any(
            (copy_json_interface, copy_quality_settings, copy_members, copy_assets, copy_labels)
        ):
            raise ValueError("At least one element has to be copied.")

        if copy_labels:
            if not copy_json_interface:
                raise ValueError(
                    "`copy_json_interface` must be set to `True` for copying the source project"
                    " labels."
                )
            if not copy_assets:
                raise ValueError(
                    "`copy_assets` must be set to `True` for copying the source project labels."
                )
            if not copy_members:
                raise ValueError(
                    "`copy_members` must be set to `True` for copying the source project labels."
                )

        fields = self.FIELDS_PROJECT
        if copy_json_interface:
            fields += self.FIELDS_JSON_INTERFACE
        if copy_quality_settings:
            fields += self.FIELDS_QUALITY_SETTINGS

        src_project = self.kili.kili_api_gateway.get_project(ProjectId(from_project_id), fields)

        if src_project["dataConnections"] and copy_assets:
            raise NotImplementedError("Copying projects with cloud storage is not supported.")

        new_project_title = title or self._generate_project_title(src_title=src_project["title"])

        new_project_description = description or ""

        json_interface = src_project["jsonInterface"] if copy_json_interface else {"jobs": {}}

        new_project_id = self.kili.create_project(
            input_type=src_project["inputType"],
            json_interface=json_interface,
            title=new_project_title,
            description=new_project_description,
        )["id"]
        logger.info(f"Creating new project with id: '{new_project_id}'")

        if copy_members:
            logger.info("Copying members...")
            self._copy_members(from_project_id, new_project_id)

        if copy_quality_settings:
            logger.info("Copying quality settings...")
            self._copy_quality_settings(new_project_id, src_project)

        if copy_assets:
            logger.info("Copying assets...")
            self._copy_assets(from_project_id, new_project_id)

        if copy_labels:
            logger.info("Copying labels...")
            self._copy_labels(from_project_id, new_project_id)

        return new_project_id

    def _generate_project_title(self, src_title: str) -> str:
        projects = self.kili.projects(
            fields=["title"], search_query=f"{src_title}%", disable_tqdm=True
        )
        proj_titles = {proj["title"] for proj in projects}
        new_title = f"{src_title} (copy)"
        i = 1
        while new_title in proj_titles:
            new_title = f"{src_title} (copy {i})"
            i += 1
        return new_title

    def _copy_members(self, from_project_id: str, new_project_id: str) -> None:
        members = self.kili.project_users(
            project_id=from_project_id,
            fields=["activated", "role", "user.email", "status", "id"],
            disable_tqdm=True,
        )

        members = [memb for memb in members if memb["status"] == "CREATED" and memb["activated"]]

        for member in tqdm(members, disable=self.disable_tqdm):
            self.kili.append_to_roles(
                project_id=new_project_id,
                user_email=member["user"]["email"],
                role=member["role"],
            )

    def _copy_quality_settings(self, new_project_id: str, src_project: Dict) -> None:
        self.kili.update_properties_in_project(
            project_id=new_project_id,
            can_skip_asset=src_project["canSkipAsset"],
            consensus_tot_coverage=src_project["consensusTotCoverage"],
            min_consensus_size=src_project["minConsensusSize"],
            use_honeypot=src_project["useHoneyPot"],
            review_coverage=src_project["reviewCoverage"],
            seconds_to_label_before_auto_assign=src_project["secondsToLabelBeforeAutoAssign"],
        )

    def _copy_assets(self, from_project_id: str, new_project_id: str) -> None:
        """Copy assets from a project to another.

        Fetches assets by batch since `content` urls expire.
        """
        filters = AssetFilters(project_id=ProjectId(from_project_id))
        options = QueryOptions(disable_tqdm=False)
        fields = (
            "content",
            "ocrMetadata",
            "externalId",
            "isHoneypot",
            "jsonContent",
            "jsonMetadata",
        )

        assets_gen = self.kili.kili_api_gateway.list_assets(filters, fields, options)

        with TemporaryDirectory() as tmp_dir:
            # TODO: modify download_media function so it can take a generator of assets
            for assets_batch in batcher(assets_gen, QUERY_BATCH_SIZE):
                downloaded_assets = self._download_assets(
                    from_project_id, fields, tmp_dir, assets_batch
                )
                self._upload_assets(new_project_id, downloaded_assets)

    def _download_assets(
        self, from_project_id: str, fields: ListOrTuple[str], tmp_dir: Path, assets: List[Dict]
    ) -> List[Dict]:
        download_function, _ = get_download_assets_function(
            self.kili.kili_api_gateway,
            download_media=True,
            fields=fields,
            project_id=ProjectId(from_project_id),
            local_media_dir=str(tmp_dir.resolve()),
        )
        assert download_function
        return download_function(assets)

    def _upload_assets(self, new_project_id: str, assets: List[Dict]) -> List[Dict]:
        # ocrMetadata field of assets need to be merged with jsonMetadata field
        for asset in assets:
            if isinstance(asset["jsonMetadata"], str):
                try:
                    asset["jsonMetadata"] = json.loads(asset["jsonMetadata"])
                except json.JSONDecodeError:
                    asset["jsonMetadata"] = {}
            if asset["ocrMetadata"]:
                asset["jsonMetadata"] = {**asset["jsonMetadata"], **asset["ocrMetadata"]}

        # we cannot send None values in the content_array or json_content_array fields of
        # kili.append_many_to_assets. So we need to sort and group the assets by the presence of
        # content and jsonContent.
        assets = sorted(
            assets,
            key=lambda asset: (bool(asset["content"]), bool(asset["jsonContent"])),
        )
        assets_iterator = itertools.groupby(
            assets,
            key=lambda asset: (bool(asset["content"]), bool(asset["jsonContent"])),
        )

        for key, group in assets_iterator:
            has_content, has_jsoncontent = key
            group = list(group)

            content_array = [asset["content"] for asset in group] if has_content else None
            external_id_array = [asset["externalId"] for asset in group]
            is_honeypot_array = [asset["isHoneypot"] for asset in group]
            json_content_array = (
                [asset["jsonContent"] for asset in group] if has_jsoncontent else None
            )
            json_metadata_array = [asset["jsonMetadata"] for asset in group]

            self.kili.append_many_to_dataset(
                project_id=new_project_id,
                content_array=content_array,
                external_id_array=external_id_array,
                is_honeypot_array=is_honeypot_array,
                json_content_array=json_content_array,
                json_metadata_array=json_metadata_array,
                disable_tqdm=True,
            )
        return assets

    # pylint: disable=too-many-locals
    def _copy_labels(self, from_project_id: str, new_project_id: str) -> None:
        assets_new_project = self.kili.kili_api_gateway.list_assets(
            AssetFilters(project_id=ProjectId(new_project_id)),
            ["id", "externalId"],
            QueryOptions(disable_tqdm=True),
        )
        assets_new_project_map = {asset["externalId"]: asset["id"] for asset in assets_new_project}

        members_new_project = self.kili.project_users(
            project_id=new_project_id, fields=["user.email", "user.id"], disable_tqdm=True
        )
        members_new_project_map = {
            member["user"]["email"]: member["user"]["id"] for member in members_new_project
        }

        nb_labels_to_copy = self.kili.kili_api_gateway.count_labels(
            LabelFilters(project_id=ProjectId(from_project_id))
        )
        if nb_labels_to_copy == 0:
            return

        labels = self.kili.labels(
            project_id=from_project_id,
            fields=[
                "author.email",
                "jsonResponse",
                "secondsToLabel",
                "isLatestLabelForUser",
                "labelOf.externalId",
                "labelType",
                "modelName",
            ],
            disable_tqdm=True,
        )
        labels = [label for label in labels if label["isLatestLabelForUser"]]

        # `append_labels` does not take arrays for `model_name` and `label_type` arguments
        # we need to sort and group the labels by `model_name` and `label_type`
        # and upload the grouped labels by batch to `append_labels`
        labels = sorted(
            labels,
            key=lambda label: (label["labelType"], label["modelName"] is None, label["modelName"]),
        )
        labels_iterator = itertools.groupby(
            labels,
            key=lambda label: (label["labelType"], label["modelName"] is None, label["modelName"]),
        )

        for key, group in labels_iterator:
            label_type, _, model_name = key
            group = list(group)

            # map external id of source project asset to
            # internal id of new project corresponding asset
            # since both source_project and new_project have the same externalIds
            asset_id_array = [
                assets_new_project_map[label["labelOf"]["externalId"]] for label in group
            ]
            json_response_array = [label["jsonResponse"] for label in group]
            author_id_array = [members_new_project_map[label["author"]["email"]] for label in group]
            seconds_to_label_array = [label["secondsToLabel"] for label in group]

            self.kili.append_labels(
                asset_id_array=asset_id_array,
                json_response_array=json_response_array,
                author_id_array=author_id_array,
                seconds_to_label_array=seconds_to_label_array,
                model_name=model_name,
                label_type=label_type,
                disable_tqdm=False,
            )



"""Client presentation methods for tags."""

from typing import Dict, List, Literal, Optional

from typeguard import typechecked

from kili.domain.project import ProjectId
from kili.domain.tag import TagId
from kili.domain.types import ListOrTuple
from kili.use_cases.tag import TagUseCases
from kili.utils.logcontext import for_all_methods, log_call

from .base import BaseClientMethods


@for_all_methods(log_call, exclude=["__init__"])
class TagClientMethods(BaseClientMethods):
    """Methods attached to the Kili client, to run actions on tags."""

    @typechecked
    def tags(
        self,
        project_id: Optional[str] = None,
        fields: ListOrTuple[str] = ("id", "organizationId", "label", "checkedForProjects"),
    ) -> List[Dict]:
        """Get tags.

        Args:
            project_id: Id of the project to which the tags belong.
                If not provided, tags of the organization are retrieved.
            fields: Fields of tags to be retrieved.
                See the [documentation](https://docs.kili-technology.com/reference/graphql-api#tag)
                for all possible fields.

        Returns:
            A list of tags as dictionaries.
        """
        tag_use_cases = TagUseCases(self.kili_api_gateway)
        return (
            tag_use_cases.get_tags_of_organization(fields=fields)
            if project_id is None
            else tag_use_cases.get_tags_of_project(project_id=ProjectId(project_id), fields=fields)
        )

    @typechecked
    def tag_project(
        self,
        project_id: str,
        tags: Optional[ListOrTuple[str]] = None,
        tag_ids: Optional[ListOrTuple[str]] = None,
        disable_tqdm: Optional[bool] = None,
    ) -> List[Dict[Literal["id"], str]]:
        """Link tags to a project.

        Args:
            project_id: Id of the project.
            tags: Sequence of tag labels to associate to the project.
            tag_ids: Sequence of tag ids to associate to the project.
                Only used if `tags` is not provided.
            disable_tqdm: Whether to disable the progress bar.

        Returns:
            A list of dictionaries with the tag ids.
        """
        tag_use_cases = TagUseCases(self.kili_api_gateway)

        if tag_ids is None:
            if tags is None:
                raise ValueError("Either `tags` or `tag_ids` must be provided.")
            tag_ids = tag_use_cases.get_tag_ids_from_labels(
                labels=tags  # pyright: ignore[reportGeneralTypeIssues]
            )

        return [
            {"id": str(tag_id)}
            for tag_id in tag_use_cases.tag_project(
                project_id=ProjectId(project_id),
                tag_ids=tag_ids,  # pyright: ignore[reportGeneralTypeIssues]
                disable_tqdm=disable_tqdm,
            )
        ]

    @typechecked
    # pylint: disable=too-many-arguments
    def untag_project(
        self,
        project_id: str,
        tags: Optional[ListOrTuple[str]] = None,
        tag_ids: Optional[ListOrTuple[str]] = None,
        all: Optional[bool] = None,  # pylint: disable=redefined-builtin
        disable_tqdm: Optional[bool] = None,
    ) -> List[Dict[Literal["id"], str]]:
        """Remove tags from a project.

        Args:
            project_id: Id of the project.
            tags: Sequence of tag labels to remove from the project.
            tag_ids: Sequence of tag ids to remove from the project.
            all: Whether to remove all tags from the project.
            disable_tqdm: Whether to disable the progress bar.

        Returns:
            A list of dictionaries with the tag ids.

        Raises:
            ValueError: Either `tags` or `tag_ids` or `all` must be provided.
        """
        if sum([tags is not None, tag_ids is not None, all is not None]) != 1:
            raise ValueError("Only one of `tags`, `tag_ids` or `all` must be provided.")

        tag_use_cases = TagUseCases(self.kili_api_gateway)

        if tag_ids is None:
            if tags is not None:
                tag_ids = tag_use_cases.get_tag_ids_from_labels(
                    labels=tags  # pyright: ignore[reportGeneralTypeIssues]
                )
            elif all is not None:
                tag_ids = [
                    tag["id"]
                    for tag in tag_use_cases.get_tags_of_project(
                        project_id=ProjectId(project_id), fields=("id",)
                    )
                ]
            else:
                raise ValueError("Either `tags` or `tag_ids` or `all` must be provided.")

        return [
            {"id": str(tag_id)}
            for tag_id in tag_use_cases.untag_project(
                project_id=ProjectId(project_id),
                tag_ids=tag_ids,  # pyright: ignore[reportGeneralTypeIssues]
                disable_tqdm=disable_tqdm,
            )
        ]

    def update_tag(self, tag_name: str, new_tag_name: str) -> Dict[Literal["id"], str]:
        """Update a tag.

        This operation is organization-wide.
        The tag will be updated for all projects of the organization.

        Args:
            tag_name: Name of the tag to update.
            new_tag_name: New name of the tag.

        Returns:
            The id of the updated tag.
        """
        tag_use_cases = TagUseCases(self.kili_api_gateway)
        tag_id = tag_use_cases.get_tag_ids_from_labels(labels=(tag_name,))[0]
        return {
            "id": str(
                tag_use_cases.update_tag(tag_id=tag_id, new_tag_name=new_tag_name).updated_tag_id
            )
        }

    def delete_tag(self, tag_name: Optional[str] = None, tag_id: Optional[str] = None) -> bool:
        """Delete the given tag.

        This operation is organization-wide.
        The tag will no longer be proposed for projects of the organization.
        If this tag is checked for one or more projects of the organization, it will be unchecked.

        Args:
            tag_name: Name of the tag to remove.
            tag_id: Id of the tag to remove.
                Use this argument if you have several tags with the same name.

        Returns:
            Whether the tag was successfully removed.
        """
        tag_use_cases = TagUseCases(self.kili_api_gateway)
        if tag_id is None:
            if tag_name is None:
                raise ValueError("Either `tag_name` or `tag_id` must be provided.")
            tag_id = tag_use_cases.get_tag_ids_from_labels(labels=(tag_name,))[0]
        return tag_use_cases.delete_tag(tag_id=TagId(tag_id))

    def create_tag(self, name: str, color: Optional[str] = None) -> Dict[Literal["id"], str]:
        """Create a tag.

        This operation is organization-wide.
        The tag will be proposed for projects of the organization.

        Args:
            name: Name of the tag to create.
            color: Color of the tag to create. If not providen a default color will be used.

        Returns:
            The id of the created tag.
        """
        tag_use_cases = TagUseCases(self.kili_api_gateway)
        return tag_use_cases.create_tag(name, color)



"""GraphQL Client."""

import logging
import os
import threading
import time
from pathlib import Path
from typing import Any, Dict, Optional, Union
from urllib.parse import urlparse

import graphql
from filelock import FileLock
from gql import Client, gql
from gql.transport import exceptions
from gql.transport.requests import RequestsHTTPTransport
from gql.transport.requests import log as gql_requests_logger
from graphql import DocumentNode, print_schema
from pyrate_limiter import Duration, Rate
from pyrate_limiter.limiter import Limiter
from tenacity import (
    retry,
    retry_all,
    retry_any,
    retry_if_exception_message,
    retry_if_exception_type,
    retry_if_not_exception_message,
    stop_after_delay,
    wait_exponential,
)

import kili.exceptions
from kili import __version__
from kili.adapters.http_client import HttpClient
from kili.core.constants import MAX_CALLS_PER_MINUTE
from kili.core.graphql.clientnames import GraphQLClientName
from kili.utils.logcontext import LogContext

gql_requests_logger.setLevel(logging.WARNING)

# _limiter and _execute_lock must be kept at module-level
# they need to be shared between all instances of Kili client within the same process

# rate limiter to avoid sending too many queries to the backend
_limiter = Limiter(Rate(MAX_CALLS_PER_MINUTE, Duration.MINUTE), max_delay=120 * 1000)

# mutex to avoid multiple threads sending queries to the backend at the same time
_execute_lock = threading.Lock()

DEFAULT_GRAPHQL_SCHEMA_CACHE_DIR = Path.home() / ".cache" / "kili" / "graphql"


# pylint: disable=too-many-instance-attributes, too-few-public-methods
class GraphQLClient:
    """GraphQL client."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        endpoint: str,
        api_key: str,
        client_name: GraphQLClientName,
        http_client: HttpClient,
        verify: Union[bool, str] = True,
        enable_schema_caching: bool = True,
        graphql_schema_cache_dir: Optional[Union[str, Path]] = DEFAULT_GRAPHQL_SCHEMA_CACHE_DIR,
    ) -> None:
        """Initialize the GraphQL client.

        Args:
            endpoint: Kili API endpoint.
            api_key: Kili API key.
            client_name: Name of the client.
            http_client: HTTP client.
            verify: Whether to verify the SSL certificate.
            enable_schema_caching: Whether to cache the GraphQL schema on disk.
            graphql_schema_cache_dir: Directory where to cache the GraphQL schema.
        """
        self.endpoint = endpoint
        self.api_key = api_key
        self.client_name = client_name
        self.http_client = http_client
        self.verify = verify
        self.enable_schema_caching = enable_schema_caching
        self.created_at = time.time()
        self.complexity_consumed = 0
        self.graphql_schema_cache_dir = (
            Path(graphql_schema_cache_dir) if graphql_schema_cache_dir else None
        )

        self.ws_endpoint = self.endpoint.replace("http", "ws")

        self._gql_transport = RequestsHTTPTransport(
            url=endpoint,
            headers=self._get_headers(),
            timeout=60,
            verify=verify,
            retries=10,
            retry_backoff_factor=0.1,  # last retry will take 0.1*2**10 = 100s
            retry_status_forcelist=(
                429,  # 429 Too Many Requests
                502,  # 502 Bad Gateway
                503,  # 503 Service Unavailable
                504,  # 504 Gateway Timeout
                520,  # Unknown Error: Generic response for an unexpected result.
                521,  # Web Server Is Down
                522,  # Connection Timed Out
            ),
        )

        if self.enable_schema_caching is True:
            if self.graphql_schema_cache_dir is None:
                raise ValueError(
                    "You must specify a cache directory if you want to enable schema caching."
                )

            self.graphql_schema_cache_dir.mkdir(parents=True, exist_ok=True)

            # mutex to avoid multiple processes operating in cache directory at the same time
            self._cache_dir_lock = FileLock(
                self.graphql_schema_cache_dir / "cache_dir.lock", timeout=15
            )

        self._gql_client = self._initizalize_graphql_client()

    def _get_headers(self) -> Dict[str, str]:
        """Get the headers."""
        return {
            "Authorization": f"X-API-Key: {self.api_key}",
            "Accept": "application/json",
            "Content-Type": "application/json",
            "apollographql-client-name": self.client_name.value,
            "apollographql-client-version": __version__,
        }

    @staticmethod
    def _get_introspection_args() -> Dict[str, bool]:
        """Get the introspection arguments."""
        return {
            "descriptions": True,  # descriptions for the schema, types, fields, and arguments
            "specified_by_url": False,  # https://spec.graphql.org/draft/#sec--specifiedBy
            "directive_is_repeatable": True,  # include repeatability of directives
            "schema_description": True,  # include schema description
            "input_value_deprecation": True,  # request deprecated input fields
        }

    def _initizalize_graphql_client(self) -> Client:
        """Initialize the GraphQL client."""
        if os.environ.get("KILI_SDK_SKIP_CHECKS", None) is not None:
            return Client(
                transport=self._gql_transport,
                fetch_schema_from_transport=False,
                introspection_args=self._get_introspection_args(),
            )

        if self.enable_schema_caching is False:
            return Client(
                transport=self._gql_transport,
                fetch_schema_from_transport=True,
                introspection_args=self._get_introspection_args(),
            )

        # In some cases (local development), we cannot get the kili version from the backend
        # and therefore we cannot determine the schema version, so we don't cache the schema
        graphql_schema_path = self._get_graphql_schema_path()
        if graphql_schema_path is None:
            return Client(
                transport=self._gql_transport,
                fetch_schema_from_transport=True,
                introspection_args=self._get_introspection_args(),
            )

        with self._cache_dir_lock:
            if not (graphql_schema_path.is_file() and graphql_schema_path.stat().st_size > 0):
                self._purge_graphql_schema_cache_dir()  # delete old schema files
                schema_str = self._get_graphql_schema_from_endpoint()
                self._cache_graphql_schema(graphql_schema_path, schema_str)

            else:
                schema_str = graphql_schema_path.read_text(encoding="utf-8")

        return Client(
            transport=self._gql_transport,
            schema=schema_str,
            introspection_args=self._get_introspection_args(),
        )

    def _get_graphql_schema_from_endpoint(self) -> str:
        """Get the GraphQL schema from the endpoint."""
        with Client(
            transport=self._gql_transport,
            fetch_schema_from_transport=True,
            introspection_args=self._get_introspection_args(),
        ) as session:
            return print_schema(session.client.schema)  # pyright: ignore[reportGeneralTypeIssues]

    def _cache_graphql_schema(self, graphql_schema_path: Path, schema_str: str) -> None:
        """Cache the graphql schema on disk."""
        with self._cache_dir_lock, graphql_schema_path.open("w", encoding="utf-8") as file:
            file.write(schema_str)
            file.flush()

    def _purge_graphql_schema_cache_dir(self) -> None:
        """Purge the schema cache directory."""
        if self.graphql_schema_cache_dir is None:
            return

        with self._cache_dir_lock:
            for file in self.graphql_schema_cache_dir.glob("*.graphql"):
                file.unlink()

    def _get_graphql_schema_path(self) -> Optional[Path]:
        """Get the path of the GraphQL schema.

        Will return None if we cannot get the schema version.
        """
        if self.graphql_schema_cache_dir is None:
            return None

        endpoint_netloc = urlparse(self.endpoint).netloc
        version = self._get_kili_app_version()
        if version is None:
            return None
        filename = f"{endpoint_netloc}_{version}.graphql"
        return self.graphql_schema_cache_dir / filename

    def _get_kili_app_version(self) -> Optional[str]:
        """Get the version of the Kili app server.

        Returns None if the version cannot be retrieved.
        """
        url = self.endpoint.replace("/graphql", "/version")
        response = self.http_client.get(url, timeout=30)
        if response.status_code == 200 and '"version":' in response.text:
            response_json = response.json()
            return response_json["version"]
        return None

    @classmethod
    def _remove_nullable_inputs(cls, variables: Dict) -> Dict:
        """Remove nullable inputs from the variables."""
        for key in ("data", "where", "project", "asset", "label", "issue"):
            if key in variables and isinstance(variables[key], dict):
                variables[key] = cls._remove_nullable_inputs(variables[key])

        return {k: v for k, v in variables.items() if v is not None}

    def execute(
        self, query: Union[str, DocumentNode], variables: Optional[Dict] = None, **kwargs
    ) -> Dict[str, Any]:
        """Execute a query.

        Args:
            query: the GraphQL query
            variables: the payload of the query
            kwargs: additional arguments to pass to the GraphQL client
        """
        document = query if isinstance(query, DocumentNode) else gql(query)
        variables = self._remove_nullable_inputs(variables) if variables else None

        try:
            return self._execute_with_retries(document, variables, **kwargs)

        except graphql.GraphQLError:  # local validation error
            # the local schema might be outdated
            # we refresh the schema and retry once
            self._purge_graphql_schema_cache_dir()
            self._gql_client = self._initizalize_graphql_client()
            try:
                return self._execute_with_retries(document, variables, **kwargs)

            except graphql.GraphQLError as err:
                # even after updating the schema, the query is invalid , we crash
                raise kili.exceptions.GraphQLError(error=err.message) from err

            # the query is valid but the server refused it, we crash
            except exceptions.TransportQueryError as err:
                raise kili.exceptions.GraphQLError(error=err.errors) from err

        except exceptions.TransportQueryError as err:  # remove validation error
            # the server refused the query after some retries, we crash
            raise kili.exceptions.GraphQLError(error=err.errors) from err

    @retry(
        reraise=True,  # re-raise the last exception
        retry=retry_all(
            retry_if_exception_type(  # error received from server
                (exceptions.TransportQueryError, exceptions.TransportServerError)
            ),
            retry_if_not_exception_message(
                match=r'.*Variable "(\$\w+)" of required type "(\w+!)" was not provided.*'
            ),
            retry_if_not_exception_message(match=r'.*Variable "(\$\w+)" got invalid value .*'),
            retry_if_not_exception_message(
                match=r'.*Field "(\w+)" is not defined by type "(\w+)".*'
            ),
            retry_any(
                retry_if_exception_message(match=r".*Invalid request made to Flagsmith API.*"),
                retry_if_exception_message(match=r".*Failed to fetch data connection.*"),
                retry_if_exception_message(match=r".*Unauthorized for url.*"),
            ),
        ),
        stop=stop_after_delay(3 * 60),
        wait=wait_exponential(multiplier=0.5, min=1, max=10),
    )
    def _execute_with_retries(
        self, document: DocumentNode, variables: Optional[Dict], **kwargs
    ) -> Dict[str, Any]:
        return self._raw_execute(document, variables, **kwargs)

    def _raw_execute(
        self, document: DocumentNode, variables: Optional[Dict], **kwargs
    ) -> Dict[str, Any]:
        _limiter.try_acquire("GraphQLClient.execute")
        with _execute_lock:
            res = self._gql_client.execute(
                document=document,
                variable_values=variables,
                extra_args={
                    "headers": {
                        **(self._gql_transport.headers or {}),
                        **LogContext(),
                    }
                },
                **kwargs,
            )
            transport = self._gql_client.transport
            if transport:
                headers = transport.response_headers  # pyright: ignore[reportGeneralTypeIssues]
                returned_complexity = int(headers.get("x-complexity", 0)) if headers else 0
                self.complexity_consumed += returned_complexity
            return res



"""Collection of Organization's related GraphQL queries and mutations."""


def get_create_organization_mutation(fragment: str) -> str:
    """Return the GraphQL createOrganization mutation."""
    return f"""
mutation(
    $data: CreateOrganizationData!
) {{
  data: createOrganization(
    data: $data
  ) {{
    {fragment}
  }}
}}
"""


def get_list_organizations_query(fragment: str) -> str:
    """Return the GraphQL organizations query."""
    return f"""
        query organizations($where: OrganizationWhere!, $first: PageSize!, $skip: Int!) {{
            data: organizations(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """


COUNT_ORGANIZATIONS_QUERY = """
        query countOrganizations($where: OrganizationWhere!) {
        data: countOrganizations(where: $where)
        }
    """


def get_get_organization_metrics_query(fragment: str) -> str:
    """Return the GraphQL organizationMetrics query."""
    return f"""
    query organizationMetrics($where: OrganizationMetricsWhere!) {{
        data: organizationMetrics(where: $where) {{
            {fragment}
        }}
    }}
    """


def get_update_properties_in_organization_mutation(fragment: str) -> str:
    """Return the GraphQL updatePropertiesInOrganization mutation."""
    return f"""
mutation(
    $data: OrganizationData!,
    $where: OrganizationWhere!
) {{
  data: updatePropertiesInOrganization(
    data: $data
    where: $where,
  ) {{
    {fragment}
  }}
}}
"""



"""Parsing of the labels of an asset."""

from typing import Dict

from kili.services.label_data_parsing.types import Project as LabelParsingProject
from kili.utils.labels.parsing import ParsedLabel, parse_labels


def parse_labels_of_asset(asset: Dict, project: LabelParsingProject) -> Dict:
    """Parse the labels of an asset queried to Kili."""
    if "labels" in asset:
        asset["labels"] = parse_labels(
            asset["labels"], project["jsonInterface"], project["inputType"]
        )
    if isinstance(asset.get("latestLabel"), Dict) and isinstance(
        asset["latestLabel"].get("jsonResponse"), Dict
    ):
        asset["latestLabel"] = ParsedLabel(
            label=asset["latestLabel"],
            json_interface=project["jsonInterface"],
            input_type=project["inputType"],
        )
    return asset



"""Develop Plugins for Kili."""

import logging
from typing import Dict, List, Optional

from kili.client import Kili
from kili.services.plugins.helpers import get_logger


class PluginCore:
    """Kili Plugin core class.

    Args:
        kili: kili instance that plugins can make use of
        project_id: the project on which plugin is ran

    Implements:

        on_submit(self, label: Dict, asset_id: str)
        on_review(self, label: Dict, asset_id: str)
        on_custom_interface_click(self, label: Dict, asset_id: str)
        on_project_updated(self, settings_updated: List[Dict])
        on_send_back_to_queue(self, asset_id: str)

    !!! warning
        if using a custom init, be sure to call super().__init__()
    """

    logger: logging.Logger
    kili: Kili
    project_id: str

    def __init__(
        self, kili: Kili, project_id: str, logger: Optional[logging.Logger] = None
    ) -> None:
        self.kili = kili
        self.project_id = project_id
        if logger:
            self.logger = logger
        else:
            self.logger = get_logger()

    def on_submit(
        self,
        label: Dict,
        asset_id: str,
    ) -> None:
        """Handler for the submit action, triggered when a default label is submitted into Kili.

        Args:
            label: Label submitted to Kili: a dictionary containing the following fields:
                `id`, `labelType`, `numberOfAnnotations`, `authorId`, `modelName`, `jsonResponse`,
                `secondsToLabel`, `isSentBackToQueue`, `search` and some technical fields:
                `createdAt`, `updatedAt`, `version`, `isLatestReviewLabelForUser`,
                `isLatestLabelForUser`, `isLatestDefaultLabelForUser`,
                `readPermissionsFromProject`.
            asset_id: Id of the asset on which the label was submitted

        !!! example
            ```python
            def on_submit(self, label: Dict, asset_id: str):
                json_response = label.get('jsonResponse')
                if label_is_respecting_business_rule(json_response):
                    return
                else:
                    self.kili.send_back_to_queue(asset_ids=[asset_id])
            ```
        """
        # pylint: disable=unused-argument
        self.logger.warning("Method not implemented. Define a custom on_submit on your plugin")

    def on_review(
        self,
        label: Dict,
        asset_id: str,
    ) -> None:
        """Handler for the review action, triggered when a default label is reviewed on Kili.

        Args:
            label: Label submitted to Kili: a dictionary containing the following fields:
                `id`, `labelType`, `numberOfAnnotations`, `authorId`, `modelName`, `jsonResponse`,
                `secondsToLabel`, `isSentBackToQueue`, `search` and `reviewedLabel` (dictionary
                that has a field `id` representing the id of the original label that was reviewed).
                It also contains some technical fields: `createdAt`, `updatedAt`, `version`,
                `isLatestReviewLabelForUser`, `isLatestLabelForUser`, `isLatestDefaultLabelForUser`,
                `readPermissionsFromProject`.
            asset_id: Id of the asset on which the label was submitted

        !!! example
            ```python
            def on_review(self, label: Dict, asset_id: str):
                json_response = label.get('jsonResponse')
                if label_is_respecting_business_rule(json_response):
                    return
                else:
                    self.kili.send_back_to_queue(asset_ids=[asset_id])
            ```
        """
        # pylint: disable=unused-argument
        self.logger.warning("Method not implemented. Define a custom on_review on your plugin")

    def on_custom_interface_click(
        self,
        label: Dict,
        asset_id: str,
    ) -> None:
        """Handler for the custom interface click action.

        !!! warning
            This handler is in beta and is still in active development,
            it should be used with caution.

        Args:
            label: Label submitted to Kili: a dictionary containing the following fields:
                `id`, `jsonResponse`.
            asset_id: id of the asset on which the action is called

        !!! example
            ```python
            def on_custom_interface_click(self, label: Dict, asset_id: str):
                json_response = label.get('jsonResponse')`
                label_id = label.get('id')
                issue = label_is_respecting_business_rule(json_response)
                if !issue:
                    return
                else:
                    self.kili.create_issues(
                            project_id=self.project_id,
                            label_id_array=[label_id],
                            text_array=[issue]
                        )
            ```
        """
        # pylint: disable=unused-argument
        self.logger.warning("Handler is in active development.")

    def on_project_updated(
        self,
        settings_updated: List[Dict],
    ) -> None:
        """Handler for the project updated action.

        Triggered when a project setting is updated on Kili.

        !!! warning
            This handler is in beta and is still in active development,
            it should be used with caution.

        Args:
            settings_updated: Settings updated on the project a list of
                dictionary containing the following fields:
                `key`, `newValue`, `oldValue`.
                !!! note
                    key is one of the following: 'canNavigateBetweenAssets',
                    'canSkipAsset', 'consensusTotCoverage', 'description',
                    'inputType', 'instructions', 'isAnonymized', 'jsonInterface',
                    'metadataTypes', 'minConsensusSize', 'reviewCoverage',
                    'title', 'archivedAt', 'useHoneyPot'

        !!! example
            ```python
            def on_project_updated(self, settings_updated: List[Dict]):
                for setting in settings_updated:
                    self.logger.info(setting)
                    # this will print:
                    # {'key': 'description', 'newValue': 'new desc', 'oldValue': 'old desc'}
            ```
        """
        # pylint: disable=unused-argument
        self.logger.warning("Handler is in active development.")

    def on_send_back_to_queue(
        self,
        asset_id: str,
    ) -> None:
        """Handler for send back to queue.

        Triggered when an asset is sent back to queue

        !!! warning
            This handler is in beta and is still in active development,
            it should be used with caution.

        Args:
            asset_id: Id of the asset on which was sent back to queue

        !!! example
            ```python
            def on_send_back_to_queue(self, asset_id: str):
                self.logger.info(f"Asset {asset_id} was sent back to queue")
            ```
        """
        # pylint: disable=unused-argument
        self.logger.warning("Handler is in active development.")



"""Helpers for the asset mutations."""

from typing import Callable, Dict, List, Optional, Union

from kili.core.helpers import convert_to_list_of_none, format_metadata, is_none_or_empty
from kili.utils.assets import PageResolution


# pylint: disable=too-many-locals
def process_update_properties_in_assets_parameters(
    asset_ids: List[str],
    *,
    external_ids: Optional[List[str]] = None,
    priorities: Optional[List[int]] = None,
    json_metadatas: Optional[List[Union[dict, str]]] = None,
    consensus_marks: Optional[List[float]] = None,
    honeypot_marks: Optional[List[float]] = None,
    to_be_labeled_by_array: Optional[List[List[str]]] = None,
    contents: Optional[List[str]] = None,
    json_contents: Optional[List[str]] = None,
    is_used_for_consensus_array: Optional[List[bool]] = None,
    is_honeypot_array: Optional[List[bool]] = None,
    resolution_array: Optional[List[dict]] = None,
    page_resolutions_array: Optional[Union[List[List[dict]], List[List[PageResolution]]]] = None,
) -> Dict:
    """Process arguments of the update_properties_in_assets method.

    Return the properties for the paginating loop.
    """
    assert asset_ids
    nb_assets_to_modify = len(asset_ids)

    input_handlers: Dict[str, Callable[[List], List]] = {
        "jsonMetadata": _handle_json_metadata,
        "shouldResetToBeLabeledBy": _handle_should_reset_to_be_labeled_by,
        "pageResolutions": _handle_page_resolutions_array,
    }

    argument_to_gql_mapping = {
        "assetId": asset_ids,
        "externalId": external_ids,
        "priority": priorities,
        "jsonMetadata": json_metadatas,
        "consensusMark": consensus_marks,
        "honeypotMark": honeypot_marks,
        "toBeLabeledBy": to_be_labeled_by_array,
        "shouldResetToBeLabeledBy": to_be_labeled_by_array,
        "content": contents,
        "jsonContent": json_contents,
        "isUsedForConsensus": is_used_for_consensus_array,
        "isHoneypot": is_honeypot_array,
        "pageResolutions": page_resolutions_array,
        "resolution": resolution_array,
    }

    return {
        k: input_handlers.get(k, lambda x: x)(convert_to_list_of_none(v, nb_assets_to_modify))
        for k, v in argument_to_gql_mapping.items()
        if v is not None
    }


# pylint: enable=too-many-locals


def _handle_page_resolutions_array(
    page_resolutions_array: Union[List[List[PageResolution]], List[List[Dict]]],
) -> List[List[Dict]]:
    page_resolutions_array_batch = []
    for page_resolution_array in page_resolutions_array:
        if page_resolution_array is None:
            page_resolutions_array_batch.append(None)
            continue

        output_page_resolution_array = []
        for page_resolution in page_resolution_array:
            output_page_resolution_array.append(
                page_resolution.as_dict()
                if isinstance(page_resolution, PageResolution)
                else page_resolution
            )
        page_resolutions_array_batch.append(output_page_resolution_array)
    return page_resolutions_array_batch


def _handle_should_reset_to_be_labeled_by(to_be_labeled_by_array):
    return list(map(is_none_or_empty, to_be_labeled_by_array))


def _handle_json_metadata(json_metadatas) -> List:
    return list(map(format_metadata, json_metadatas))



"""GraphQL Issues operations."""

GQL_CREATE_ISSUES = """
mutation createIssues(
    $issues: [IssueToCreate!]!
    $where: AssetWhere!
) {
  data: createIssues(
    issues: $issues
    where: $where
  ) {
    id
  }
}
"""

GQL_COUNT_ISSUES = """
query countIssues($where: IssueWhere!) {
    data: countIssues(where: $where)
}
"""

GQL_UPDATE_ISSUE = """
mutation UpdatePropertiesInIssue($data: IssueData!, $where: IssueWhere!) {
  updatePropertiesInIssue(data: $data, where: $where) {
    id
  }
}
"""


def get_issues_query(fragment: str) -> str:
    """Return the GraphQL issues query."""
    return f"""
    query issues($where: IssueWhere!, $first: PageSize!, $skip: Int!) {{
        data: issues(where: $where, first: $first, skip: $skip) {{
            {fragment}
        }}
    }}
    """



"""User domain."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Literal, NewType, Optional

if TYPE_CHECKING:
    from .organization import OrganizationId

UserId = NewType("UserId", str)

HubspotSubscriptionStatus = Literal["SUBSCRIBED", "UNSUBSCRIBED"]


@dataclass
class UserFilter:
    """User filters for running a users search."""

    id: Optional[UserId]
    activated: Optional[bool] = None
    api_key: Optional[str] = None
    email: Optional[str] = None
    id_in: Optional[List[UserId]] = None
    organization_id: Optional["OrganizationId"] = None



"""Common code for the llm exporter."""

import json
import logging
import warnings
from ast import literal_eval
from pathlib import Path
from typing import Dict, List, Optional, Union

from kili.services.asset_import.helpers import SEPARATOR
from kili.services.export.exceptions import NotCompatibleInputType
from kili.services.export.format.base import AbstractExporter
from kili.services.export.format.llm.types import ExportLLMItem, RankingValue
from kili.services.types import Job


class LLMExporter(AbstractExporter):
    """Common code for LLM exporters."""

    def _check_arguments_compatibility(self) -> None:
        """Checks if the export label format is compatible with the export options."""
        if not self.with_assets:
            raise ValueError("LLM export format requires assets to be downloaded.")

    def _check_project_compatibility(self) -> None:
        """Checks if the export label format is compatible with the project."""
        if self.project["inputType"] != "LLM_RLHF":
            raise NotCompatibleInputType(
                f"Project with input type \"{self.project['inputType']}\" not compatible with LLM"
                " export format."
            )

    def _is_job_compatible(self, job: Job) -> bool:
        """Check job compatibility with the LLM format."""
        _ = job
        return True

    def _save_assets_export(self, assets: List[Dict], output_filename: Path) -> None:
        """Save the assets to a file and return the link to that file."""
        self.logger.info("Exporting to llm format...")

        export_json = json.dumps(assets, sort_keys=False, indent=4)
        with output_filename.open("wb") as output_file:
            output_file.write(export_json.encode("utf-8"))

    def process_and_save(
        self, assets: List[Dict], output_filename: Path
    ) -> Optional[List[Dict[str, Union[List[str], str]]]]:
        """LLM specific process and save."""
        result = self.process(assets)
        self._save_assets_export(result, output_filename)

    def process(self, assets: List[Dict]) -> List[Dict[str, Union[List[str], str]]]:
        """LLM specific process."""
        warnings.warn(
            "Exporting llm labels with `kili.export_labels` is deprecated."
            " Please use `kili.llm.export` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        clean_assets = self.preprocess_assets(assets)
        if self.label_format == "llm_v1":
            return self._process_llm_v1(clean_assets)
        return self._process_llm_dynamic_v1(clean_assets)

    def _process_llm_dynamic_v1(self, assets: List[Dict]) -> List[Dict[str, Union[List[str], str]]]:
        result = []
        for asset in assets:
            step_number = _count_step(asset)
            label = asset["latestLabel"]
            steps = {}
            context = []
            formatted_asset = _format_raw_data(asset, all_model_keys=True)
            for i in range(step_number):
                steps[f"{i}"] = {
                    "raw_data": context + _format_raw_data(asset, i),
                    "status": asset["status"],
                    "external_id": asset["externalId"],
                    "metadata": asset["jsonMetadata"],
                    "labels": [
                        {
                            "author": label["author"]["email"],
                            "created_at": label["createdAt"],
                            "label_type": label["labelType"],
                            "label": _format_json_response_dynamic(
                                self.project["jsonInterface"]["jobs"], label["jsonResponse"], i
                            ),
                        }
                    ],
                }
                next_context = _get_next_step_context(formatted_asset, label["jsonResponse"], i)
                context = context + next_context

            if step_number > 0:
                result.append(steps)
        return result

    def _process_llm_v1(self, assets: List[Dict]) -> List[Dict[str, Union[List[str], str]]]:
        result = []
        if len(assets) == 0:
            return result
        for asset in assets:
            result.append(
                {
                    "raw_data": _format_raw_data(asset),
                    "status": asset["status"],
                    "external_id": asset["externalId"],
                    "metadata": asset["jsonMetadata"],
                    "labels": list(
                        map(
                            lambda label: {
                                "author": label["author"]["email"],
                                "created_at": label["createdAt"],
                                "label_type": label["labelType"],
                                "label": _format_json_response(
                                    self.project["jsonInterface"]["jobs"], label["jsonResponse"]
                                ),
                            },
                            asset["labels"],
                        )
                    ),
                }
            )
        return result


def _get_step_ranking_value(json_response: Dict, step_number: int) -> RankingValue:
    prefix = f"STEP_{step_number+1}_"
    for category in json_response["CLASSIFICATION_JOB"]["categories"]:
        if category["name"] != f"STEP_{step_number+1}":
            continue

        for children_name, children_value in category["children"].items():
            if children_name == f"STEP_{step_number+1}_RANKING":
                raw_value = children_value["categories"][0]["name"]
                return raw_value[len(prefix) :]
    return RankingValue.TIE


def _get_next_step_context(
    formatted_asset: List[ExportLLMItem], json_response: Dict, step_number: int
) -> List[ExportLLMItem]:
    context = []
    skipped_context = 0
    completion_index = 0
    ranking = _get_step_ranking_value(json_response, step_number)
    for item in formatted_asset:
        if skipped_context > step_number:
            break

        if skipped_context == step_number:
            if item["role"] == "user":
                context.append(item)
            else:
                if completion_index == 0 and ranking in ["A_1", "A_2", "A_3", "TIE"]:
                    context.append(item)
                    break
                if completion_index == 1 and ranking in ["B_1", "B_2", "B_3"]:
                    context.append(item)
                    break
                completion_index += 1

        if item["role"] == "assistant":
            skipped_context += 1

    return context


def _count_step(asset: Dict) -> int:
    label = asset["latestLabel"]
    if "jsonResponse" not in label and "CLASSIFICATION_JOB" not in label["jsonResponse"]:
        return 0
    return len(label["jsonResponse"]["CLASSIFICATION_JOB"]["categories"])


def _format_json_response_dynamic(
    jobs_config: Dict, json_response: Dict, step_number: int
) -> Dict[str, Union[str, List[str]]]:
    # check subjobs of the step
    job_step = f"STEP_{step_number+1}"
    for item in json_response["CLASSIFICATION_JOB"]["categories"]:
        if item["name"] != job_step:
            continue
        response_step = _format_json_response(jobs_config, item["children"])
        formatted_response = literal_eval(str(response_step).replace(job_step + "_", ""))
        return formatted_response
    return {}


def _format_json_response(
    jobs_config: Dict, json_response: Dict
) -> Dict[str, Union[str, List[str]]]:
    result = {}
    for job_name, job_value in json_response.items():
        job_config = jobs_config[job_name]
        if job_config is None:
            continue
        if job_config["mlTask"] == "CLASSIFICATION":
            result[job_name] = []
            for category in job_value["categories"]:
                result[job_name].append(category["name"])
                if "children" in category:
                    for child_name, child_value in _format_json_response(
                        jobs_config, category["children"]
                    ).items():
                        result[f"{job_name}.{category['name']}.{child_name}"] = child_value
        elif job_config["mlTask"] == "TRANSCRIPTION":
            result[job_name] = job_value["text"]
        else:
            logging.warning(f"Job {job_name} with mlTask {job_config['mlTask']} not supported")
    return result


def _format_raw_data(
    asset, step_number: Optional[int] = None, all_model_keys: Optional[bool] = False
) -> List[ExportLLMItem]:
    raw_data = []

    chat_id = asset["jsonMetadata"].get("chat_id", None)

    if (
        "chat_item_ids" in asset["jsonMetadata"]
        and isinstance(asset["jsonMetadata"]["chat_item_ids"], str)
        and len(asset["jsonMetadata"]["chat_item_ids"]) > 0
    ):
        chat_items_ids = str.split(asset["jsonMetadata"]["chat_item_ids"], SEPARATOR)
        if step_number is not None:
            chat_items_ids = chat_items_ids[step_number * 3 :]
    else:
        chat_items_ids = []

    if (
        "models" in asset["jsonMetadata"]
        and isinstance(asset["jsonMetadata"]["models"], str)
        and len(asset["jsonMetadata"]["models"]) > 0
    ):
        models = str.split(asset["jsonMetadata"]["models"], SEPARATOR)
    else:
        models = []

    with Path(asset["content"]).open(encoding="utf8") as file:
        data = json.load(file)
    version = data.get("version", None)
    if version == "0.1":
        prompts = data["prompts"]
        if step_number is not None:
            prompts = [prompts[step_number]]
        for index, prompt in enumerate(prompts):
            raw_data.append(
                ExportLLMItem(
                    {
                        "role": prompt.get("title", "user"),
                        "content": prompt["prompt"],
                        "id": _safe_pop(chat_items_ids),
                        "chat_id": chat_id,
                        "model": None,
                    }
                )
            )
            raw_data.extend(
                ExportLLMItem(
                    {
                        "role": completion.get("title", "assistant"),
                        "content": completion["content"],
                        "id": _safe_pop(chat_items_ids),
                        "chat_id": chat_id,
                        "model": models[index_completion]
                        if (
                            (index == len(prompts) - 1 or all_model_keys)
                            and len(models) > index_completion
                        )
                        else None,
                    }
                )
                for index_completion, completion in enumerate(prompt["completions"])
            )
    else:
        raise ValueError(f"Version {version} not supported")
    return raw_data


def _safe_pop(lst, index=0):
    try:
        return lst.pop(index)
    except IndexError:
        return None



"""Utilities to handle assets."""

from dataclasses import dataclass
from typing import Any, Dict


@dataclass
class PageResolution:
    """A wrapper for PageResolution GraphQL object."""

    width: float
    height: float
    page_number: int
    rotation: int = 0

    def as_dict(self) -> Dict[str, Any]:
        """Return a dictionary representation of the object."""
        output = {
            "width": self.width,
            "height": self.height,
            "pageNumber": self.page_number,
        }
        if self.rotation != 0:
            output["rotation"] = self.rotation

        return output



"""Asset domain."""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, NewType, Optional

from kili.domain.types import ListOrTuple

if TYPE_CHECKING:
    from kili.domain.issue import IssueStatus, IssueType
    from kili.domain.label import LabelType
    from kili.domain.project import ProjectId

AssetId = NewType("AssetId", str)
AssetExternalId = NewType("AssetExternalId", str)


AssetStatus = Literal["TODO", "ONGOING", "LABELED", "REVIEWED", "TO_REVIEW"]


@dataclass
class AssetFilters:
    """Asset filters for running an asset search."""

    # pylint: disable=too-many-instance-attributes
    project_id: "ProjectId"
    asset_id: Optional[AssetId] = None
    asset_id_in: Optional[ListOrTuple[AssetId]] = None
    asset_id_not_in: Optional[ListOrTuple[AssetId]] = None
    consensus_mark_gte: Optional[float] = None
    consensus_mark_lte: Optional[float] = None
    external_id_strictly_in: Optional[ListOrTuple[AssetExternalId]] = None
    external_id_in: Optional[ListOrTuple[AssetExternalId]] = None
    honeypot_mark_gte: Optional[float] = None
    honeypot_mark_lte: Optional[float] = None
    label_author_in: Optional[ListOrTuple[str]] = None
    label_consensus_mark_gte: Optional[float] = None
    label_consensus_mark_lte: Optional[float] = None
    label_created_at: Optional[str] = None
    label_created_at_gte: Optional[str] = None
    label_created_at_lte: Optional[str] = None
    label_honeypot_mark_gte: Optional[float] = None
    label_honeypot_mark_lte: Optional[float] = None
    label_type_in: Optional[ListOrTuple["LabelType"]] = None
    label_labeler_in: Optional[ListOrTuple[str]] = None
    label_labeler_not_in: Optional[ListOrTuple[str]] = None
    label_reviewer_in: Optional[ListOrTuple[str]] = None
    label_reviewer_not_in: Optional[ListOrTuple[str]] = None
    assignee_in: Optional[ListOrTuple[str]] = None
    assignee_not_in: Optional[ListOrTuple[str]] = None
    metadata_where: Optional[dict] = None
    skipped: Optional[bool] = None
    status_in: Optional[ListOrTuple[AssetStatus]] = None
    updated_at_gte: Optional[str] = None
    updated_at_lte: Optional[str] = None
    label_category_search: Optional[str] = None
    created_at_gte: Optional[str] = None
    created_at_lte: Optional[str] = None
    inference_mark_gte: Optional[float] = None
    inference_mark_lte: Optional[float] = None
    issue_type: Optional["IssueType"] = None
    issue_status: Optional["IssueStatus"] = None



"""Service for exporting kili objects."""

from typing import Dict, List, Optional, Union

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.domain.asset.asset import AssetFilters
from kili.domain.label import LabelType
from kili.domain.project import ProjectId

from .dynamic import LLMDynamicExporter
from .static import LLMStaticExporter

CHAT_ITEMS_NEEDED_FIELDS = [
    "id",
    "content",
    "createdAt",
    "modelId",
    "parentId",
    "role",
]

LABELS_NEEDED_FIELDS = [
    "annotations.id",
    "author.id",
    "author.email",
    "author.firstname",
    "author.lastname",
    *(f"chatItems.{field}" for field in CHAT_ITEMS_NEEDED_FIELDS),
    "createdAt",
    "id",
    "isLatestLabelForUser",
    "isSentBackToQueue",
    "jsonResponse",  # This is needed to keep annotations
    "labelType",
    "modelName",
]

ASSET_DYNAMIC_NEEDED_FIELDS = [
    "assetProjectModels.id",
    "assetProjectModels.configuration",
    "assetProjectModels.projectModelId",
    "content",
    "externalId",
    "jsonMetadata",
    *(f"labels.{field}" for field in LABELS_NEEDED_FIELDS),
    "status",
]

ASSET_STATIC_NEEDED_FIELDS = [
    "content",
    "externalId",
    "jsonMetadata",
    "labels.jsonResponse",
    "labels.author.id",
    "labels.author.email",
    "labels.author.firstname",
    "labels.author.lastname",
    "labels.createdAt",
    "labels.isLatestLabelForUser",
    "labels.isSentBackToQueue",
    "labels.labelType",
    "labels.modelName",
    "status",
]


def export(  # pylint: disable=too-many-arguments, too-many-locals
    kili_api_gateway: KiliAPIGateway,
    project_id: ProjectId,
    asset_filter: AssetFilters,
    disable_tqdm: Optional[bool],
    include_sent_back_labels: Optional[bool],
    label_type_in: List[LabelType],
) -> Optional[List[Dict[str, Union[List[str], str]]]]:
    """Export the selected assets with their labels into the required format, and save it into a file archive."""
    project = kili_api_gateway.get_project(project_id, ["id", "inputType", "jsonInterface"])
    input_type = project["inputType"]

    fields = get_fields_to_fetch(input_type)
    assets = list(
        kili_api_gateway.list_assets(asset_filter, fields, QueryOptions(disable_tqdm=disable_tqdm))
    )
    cleaned_assets = preprocess_assets(assets, include_sent_back_labels or False, label_type_in)

    if input_type == "LLM_RLHF":
        return LLMStaticExporter(kili_api_gateway).export(
            cleaned_assets, project_id, project["jsonInterface"]
        )
    if input_type == "LLM_INSTR_FOLLOWING":
        return LLMDynamicExporter(kili_api_gateway).export(cleaned_assets, project["jsonInterface"])
    raise ValueError(f'Project Input type "{input_type}" cannot be used for llm exports.')


def get_fields_to_fetch(input_type: str) -> List[str]:
    """Return the fields to fetch depending on the export type."""
    if input_type == "LLM_RLHF":
        return ASSET_STATIC_NEEDED_FIELDS
    return ASSET_DYNAMIC_NEEDED_FIELDS


def preprocess_assets(
    assets: List[Dict], include_sent_back_labels: bool, label_type_in: List[LabelType]
) -> List[Dict]:
    """Format labels in the requested format, and filter out autosave labels."""
    assets_in_format = []
    for asset in assets:
        if "labels" in asset:
            labels_of_asset = []
            for label in asset["labels"]:
                labels_of_asset.append(label)
                labels_of_asset = list(
                    filter(lambda label: label["labelType"] in label_type_in, labels_of_asset)
                )
                if not include_sent_back_labels:
                    labels_of_asset = list(
                        filter(lambda label: label["isSentBackToQueue"] is False, labels_of_asset)
                    )
            if len(labels_of_asset) > 0:
                asset["labels"] = labels_of_asset
                assets_in_format.append(asset)
    return assets_in_format



"""Tools for export with videos."""

import os
import shutil
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import ffmpeg

from kili.services.export.exceptions import NotExportableAssetError
from kili.utils.tempfile import TemporaryDirectory


class FFmpegError(Exception):
    """Errors related to ffmpeg."""


def get_video_dimensions(asset: Dict) -> Tuple:
    """Get a video width and height."""
    if "resolution" in asset and asset["resolution"] is not None:
        return (asset["resolution"]["width"], asset["resolution"]["height"])

    if Path(asset["content"]).is_file():
        probe = ffmpeg.probe(str(asset["content"]))
        video_info = next(s for s in probe["streams"] if s["codec_type"] == "video")
        width = video_info["width"]
        height = video_info["height"]
        return width, height

    raise NotExportableAssetError(
        f"Could not find video dimensions for asset with externalId '{asset['externalId']}'. Please"
        " use `kili.update_properties_in_assets()` to update the resolution of your asset. Or use"
        " `kili.export_labels(with_assets=True).`"
    )


def cut_video(
    video_path: Path, asset: Dict, leading_zeros: int, output_dir: Optional[Path]
) -> List[Path]:
    """Download and cut video into frames."""
    output_dir = output_dir or video_path.parent
    output_dir.mkdir(parents=True, exist_ok=True)

    if (
        "jsonMetadata" in asset
        and "processingParameters" in asset["jsonMetadata"]
        and "framesPlayedPerSecond" in asset["jsonMetadata"]["processingParameters"]
    ):
        metadata = asset["jsonMetadata"]
        final_framerate = metadata["processingParameters"]["framesPlayedPerSecond"]
    else:
        try:
            probe = ffmpeg.probe(str(video_path))
            video_info = next(s for s in probe["streams"] if s["codec_type"] == "video")
            # avg_frame_rate is total # of frames / total duration
            # r_frame_rate is the lowest framerate with which all timestamps can be represented
            # accurately (it is the least common multiple of all framerates in the stream)
            frame_rate_string = video_info["r_frame_rate"].split("/")
            final_framerate = int(frame_rate_string[0]) / int(frame_rate_string[1])
        except ffmpeg.Error as error:
            raise FFmpegError(f"ffmpeg error for asset {video_path.name}: {error}") from error

    output_frames = []
    try:
        with TemporaryDirectory() as temp_dir:
            ffmpeg.input(str(video_path)).filter("fps", fps=final_framerate, round="up").output(
                os.path.join(str(temp_dir), "%d.jpg"), start_number=0
            ).run(capture_stdout=True, capture_stderr=True)
            for file in os.listdir(temp_dir):
                idx = int(Path(file).stem)
                new_filename = f'{asset["externalId"]}_{str(idx+1).zfill(leading_zeros)}.jpg'
                shutil.move(str(temp_dir / file), str(output_dir / new_filename))
                output_frames.append(output_dir / new_filename)
    except ffmpeg.Error as error:
        raise FFmpegError(f"ffmpeg error for asset {video_path.name}: {error}") from error

    return sorted(output_frames)



"""Handle LLM_INSTR_FOLLOWING project exports."""

import logging
from typing import Dict, List, Union

from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway

CHAT_ITEMS_NEEDED_FIELDS = [
    "id",
    "content",
    "modelId",
    "parentId",
    "role",
]

LABELS_NEEDED_FIELDS = [
    "annotations.id",
    "author.id",
    "author.email",
    "author.firstname",
    "author.lastname",
    *(f"chatItems.{field}" for field in CHAT_ITEMS_NEEDED_FIELDS),
    "createdAt",
    "id",
    "isLatestLabelForUser",
    "jsonResponse",  # This is needed to keep annotations
    "labelType",
    "modelName",
]


class LLMDynamicExporter:
    """Handle exports of LLM_RLHF projects."""

    def __init__(self, kili_api_gateway: KiliAPIGateway):
        self.kili_api_gateway = kili_api_gateway

    def export(
        self, assets: List[Dict], json_interface: Dict
    ) -> List[Dict[str, Union[List[str], str]]]:
        """Asset content depends of each label."""
        export_res = []
        for asset in assets:
            # obfuscate models here
            obfuscated_models = {}
            for index, asset_project_model in enumerate(asset["assetProjectModels"]):
                obfuscated_models[asset_project_model["id"]] = f"{chr(65 + index)}"
            for label in asset["labels"]:
                result = {}
                chat_items = label["chatItems"]
                annotations = label["annotations"]
                rounds = self._build_rounds(chat_items, annotations, json_interface)
                total_rounds = len(rounds)
                for step, round in enumerate(rounds):
                    raw_data = _format_raw_data(
                        round["context"]
                        + round["pre_prompts"]
                        + [round["prompt"]]
                        + round["completion"],
                        label["id"],
                        obfuscated_models,
                    )
                    formatted_response = _format_json_response(
                        json_interface["jobs"],
                        round["annotations"],
                        round["completion"],
                        obfuscated_models,
                    )
                    label_data = {
                        "author": label["author"]["email"],
                        "created_at": label["createdAt"],
                        "label_type": label["labelType"],
                        "label": {},
                    }
                    if formatted_response["turn"]:
                        label_data["label"]["turn"] = formatted_response["turn"]
                    if step == total_rounds - 1 and formatted_response["conversation"]:
                        label_data["label"]["conversation"] = formatted_response["conversation"]

                    result[f"{step}"] = {
                        "external_id": asset["externalId"],
                        "metadata": asset["jsonMetadata"],
                        "models": _format_models_object(
                            asset["assetProjectModels"], obfuscated_models
                        ),
                        "labels": [label_data],
                        "raw_data": raw_data,
                        "status": asset["status"],
                    }
                export_res.append(result)
        return export_res

    def _get_round_winner(self, completions, annotations, json_interface):
        """Get the winner completion.

        Find the first comparison job in the json_interface and return the completion that has an annotation of this job targeting it.
        """
        comparison_job = None
        comparison_job_name = None
        for job_name, job in json_interface["jobs"].items():
            if job["mlTask"] == "COMPARISON":
                comparison_job = job
                comparison_job_name = job_name
                break
        if comparison_job is None:
            raise ValueError("No comparison job found in json_interface")

        winnerId = None
        for annotation in annotations:
            if annotation["job"] == comparison_job_name:
                winnerId = annotation["annotationValue"]["choice"]["firstId"]

        if winnerId is None:
            raise ValueError(
                f"Job {comparison_job_name} is used to create context, but no related annotations found"
            )
        res = list(filter(lambda completion: completion["id"] == winnerId, completions))
        return res[0] if len(res) > 0 else None

    def _init_round(self, context):
        return {
            "context": context,
            "prompt": None,
            "pre_prompts": [],
            "completion": [],
            "annotations": [],
        }

    def _build_rounds(self, chat_items, annotations, json_interface):
        """A round is composed of a prompt with n pre-prompts and n completions."""
        dict_chat_items = {}
        for chat_item in chat_items:
            if dict_chat_items.get(chat_item["parentId"]) is None:
                dict_chat_items[chat_item["parentId"]] = []
            dict_chat_items[chat_item["parentId"]].append(chat_item)
        rounds = []
        parent_target = None
        has_children = True
        current_round = self._init_round([])

        while has_children:
            nodes = dict_chat_items.get(parent_target)
            if nodes is None or len(nodes) == 0:
                has_children = False
                continue
            node = nodes[0]
            if node["role"].lower() == "system":
                current_round["pre_prompts"].append(node)
                parent_target = node["id"]
                current_round["annotations"] += [
                    annotation
                    for annotation in annotations
                    if annotation["chatItemId"] == node["id"]
                ]
                continue

            if node["role"].lower() == "user":
                current_round["prompt"] = node
                parent_target = node["id"]
                current_round["annotations"] += [
                    annotation
                    for annotation in annotations
                    if annotation["chatItemId"] == node["id"]
                ]
                continue

            if node["role"].lower() == "assistant":
                has_children = False
                if dict_chat_items.get(parent_target) is None:
                    continue
                for chat_item in dict_chat_items[parent_target]:
                    current_round["completion"].append(chat_item)
                    current_round["annotations"] += [
                        annotation
                        for annotation in annotations
                        if annotation["chatItemId"] == chat_item["id"]
                    ]
                    if not has_children and dict_chat_items.get(chat_item["id"]) is not None:
                        has_children = True
                        parent_target = chat_item["id"]

                current_round["annotations"] += [
                    annotation for annotation in annotations if annotation["chatItemId"] is None
                ]
                rounds.append(current_round)
                new_context = (
                    current_round["context"]
                    + current_round["pre_prompts"]
                    + [
                        current_round["prompt"],
                        self._get_round_winner(
                            current_round["completion"],
                            current_round["annotations"],
                            json_interface,
                        ),
                    ]
                )
                current_round = self._init_round(new_context)
                continue

            raise ValueError(f"Role {node['role']} not supported")
        if current_round["prompt"] is not None:
            current_round["annotations"] += [
                annotation for annotation in annotations if annotation["chatItemId"] is None
            ]
            rounds.append(current_round)
        return rounds


def _format_classification_annotation(annotation):
    return annotation["annotationValue"]["categories"]


def _format_transcription_annotation(annotation):
    return annotation["annotationValue"]["text"]


def _format_comparison_annotation(annotation, completions, job, obfuscated_models):
    """Return A_X or B_X depending of the evaluation completion and its score."""
    model_id = None
    for completion in completions:
        if annotation["annotationValue"]["choice"]["firstId"] == completion["id"]:
            model_id = completion["modelId"]
            break

    if model_id is None:
        raise ValueError(f"Failed to found model of annotation {annotation['id']}")

    iteration = 0
    for _comparison_code, comparison_note in job["content"]["options"].items():
        iteration += 1
        if comparison_note["name"] == annotation["annotationValue"]["choice"]["code"]:
            break

    return f"{obfuscated_models[model_id]}_{iteration}"


def _format_json_response(
    jobs_config: Dict, annotations: List[Dict], completions: List[Dict], obfuscated_models: Dict
) -> Dict[str, Dict[str, Union[str, List[str]]]]:
    result = {"turn": {}, "conversation": {}}
    for annotation in annotations:
        formatted_response = None
        job = jobs_config[annotation["job"]]
        if job["mlTask"] == "CLASSIFICATION":
            formatted_response = _format_classification_annotation(annotation)
        elif job["mlTask"] == "TRANSCRIPTION":
            formatted_response = _format_transcription_annotation(annotation)
        elif job["mlTask"] == "COMPARISON":
            formatted_response = _format_comparison_annotation(
                annotation, completions, job, obfuscated_models
            )

        if formatted_response is None:
            logging.warning(
                f"Annotation with job {annotation['job']} with mlTask {job['mlTask']} not supported. Ignored in the export."
            )
        elif "level" in job and job["level"] == "conversation":
            result["conversation"][annotation["job"]] = formatted_response
        else:
            result["turn"][annotation["job"]] = formatted_response

    return result


def _format_raw_data(chat_items, label_id, obfuscated_models):
    raw_data = []
    for chat_item in chat_items:
        raw_data.append(
            {
                "content": chat_item["content"],
                "role": chat_item["role"].lower(),
                "chat_id": label_id,
                "id": chat_item["id"],
                "model": obfuscated_models[chat_item["modelId"]]
                if chat_item["modelId"] in obfuscated_models
                else None,
            }
        )
    return raw_data


def _format_models_object(models, obfuscated_models):
    res = {}
    for model in models:
        model_configuration = {
            "id": model["id"],
            "projectModelId": model["projectModelId"],
            "configuration": model["configuration"],
        }
        res[obfuscated_models[model["id"]]] = model_configuration

    return res



"""Client presentation methods for projects."""

from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    cast,
    overload,
)

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.enums import ProjectType
from kili.domain.project import ComplianceTag, InputType, ProjectFilters, ProjectId
from kili.domain.tag import TagId
from kili.domain.types import ListOrTuple
from kili.presentation.client.helpers.common_validators import (
    disable_tqdm_if_as_generator,
)
from kili.use_cases.project.project import ProjectUseCases
from kili.use_cases.tag import TagUseCases
from kili.utils.logcontext import for_all_methods, log_call

from .base import BaseClientMethods


@for_all_methods(log_call, exclude=["__init__"])
class ProjectClientMethods(BaseClientMethods):
    """Methods attached to the Kili client, to run actions on projects."""

    @typechecked
    # pylint: disable=too-many-arguments
    def create_project(
        self,
        title: str,
        description: str = "",
        input_type: Optional[InputType] = None,
        json_interface: Optional[Dict] = None,
        project_id: Optional[ProjectId] = None,
        project_type: Optional[ProjectType] = None,
        tags: Optional[ListOrTuple[str]] = None,
        compliance_tags: Optional[ListOrTuple[ComplianceTag]] = None,
    ) -> Dict[Literal["id"], str]:
        # pylint: disable=line-too-long
        """Create a project.

        Args:
            input_type: Currently, one of `IMAGE`, `PDF`, `TEXT` or `VIDEO`.
            json_interface: The json parameters of the project, see Edit your interface.
            title: Title of the project.
            description: Description of the project.
            project_id: Identifier of the project to copy.
            project_type: Currently, one of:

                - `IMAGE_CLASSIFICATION_MULTI`
                - `IMAGE_CLASSIFICATION_SINGLE`
                - `IMAGE_OBJECT_DETECTION_POLYGON`
                - `IMAGE_OBJECT_DETECTION_RECTANGLE`
                - `IMAGE_OBJECT_DETECTION_SEMANTIC`
                - `IMAGE_POSE_ESTIMATION`
                - `OCR`
                - `PDF_CLASSIFICATION_MULTI`
                - `PDF_CLASSIFICATION_SINGLE`
                - `PDF_NAMED_ENTITY_RECOGNITION`
                - `PDF_OBJECT_DETECTION_RECTANGLE`
                - `SPEECH_TO_TEXT`
                - `TEXT_CLASSIFICATION_MULTI`
                - `TEXT_CLASSIFICATION_SINGLE`
                - `TEXT_NER`
                - `TEXT_TRANSCRIPTION`
                - `TIME_SERIES`
                - `VIDEO_CLASSIFICATION_SINGLE`
                - `VIDEO_FRAME_CLASSIFICATION`
                - `VIDEO_FRAME_OBJECT_TRACKING`

            tags: Tags to add to the project. The tags must already exist in the organization.
            compliance_tags: Compliance tags of the project.
                Compliance tags are used to categorize projects based on the sensitivity of
                the data being handled and the legal constraints associated with it.
                Possible values are: `PHI` and `PII`.

        Returns:
            A dict with the id of the created project.

        Examples:
            >>> kili.create_project(input_type='IMAGE', json_interface=json_interface, title='Example')

        !!! example "Recipe"
            For more detailed examples on how to create projects,
                see [the recipe](https://docs.kili-technology.com/recipes/creating-a-project).
        """
        project_id = ProjectUseCases(self.kili_api_gateway).create_project(
            input_type=input_type,
            json_interface=json_interface,
            title=title,
            description=description,
            project_id=project_id,
            project_type=project_type,
            compliance_tags=compliance_tags,
        )

        if tags is not None:
            tag_use_cases = TagUseCases(self.kili_api_gateway)
            tag_ids = tag_use_cases.get_tag_ids_from_labels(labels=tags)
            tag_use_cases.tag_project(
                project_id=project_id, tag_ids=cast(ListOrTuple[TagId], tag_ids), disable_tqdm=True
            )

        return {"id": project_id}

    @overload
    # pylint: disable=too-many-arguments
    def projects(
        self,
        project_id: Optional[str] = None,
        search_query: Optional[str] = None,
        should_relaunch_kpi_computation: Optional[bool] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        archived: Optional[bool] = None,
        starred: Optional[bool] = None,
        tags_in: Optional[ListOrTuple[str]] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "consensusTotCoverage",
            "id",
            "inputType",
            "jsonInterface",
            "minConsensusSize",
            "reviewCoverage",
            "roles.id",
            "roles.role",
            "roles.user.email",
            "roles.user.id",
            "title",
        ),
        deleted: Optional[bool] = None,
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    # pylint: disable=too-many-arguments
    def projects(
        self,
        project_id: Optional[str] = None,
        search_query: Optional[str] = None,
        should_relaunch_kpi_computation: Optional[bool] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        archived: Optional[bool] = None,
        starred: Optional[bool] = None,
        tags_in: Optional[ListOrTuple[str]] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "consensusTotCoverage",
            "id",
            "inputType",
            "jsonInterface",
            "minConsensusSize",
            "reviewCoverage",
            "roles.id",
            "roles.role",
            "roles.user.email",
            "roles.user.id",
            "title",
        ),
        deleted: Optional[bool] = None,
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    # pylint: disable=too-many-arguments,too-many-locals
    def projects(
        self,
        project_id: Optional[str] = None,
        search_query: Optional[str] = None,
        should_relaunch_kpi_computation: Optional[bool] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        archived: Optional[bool] = None,
        starred: Optional[bool] = None,
        tags_in: Optional[ListOrTuple[str]] = None,
        organization_id: Optional[str] = None,
        fields: ListOrTuple[str] = (
            "consensusTotCoverage",
            "id",
            "inputType",
            "jsonInterface",
            "minConsensusSize",
            "reviewCoverage",
            "roles.id",
            "roles.role",
            "roles.user.email",
            "roles.user.id",
            "title",
        ),
        deleted: Optional[bool] = None,
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of projects that match a set of criteria.

        Args:
            project_id: Select a specific project through its project_id.
            search_query: Returned projects with a title or a description matching this [PostgreSQL ILIKE](https://www.postgresql.org/docs/current/functions-matching.html#FUNCTIONS-LIKE) pattern.
            should_relaunch_kpi_computation: Deprecated, do not use.
            updated_at_gte: Returned projects should have a label whose update date is greater or equal
                to this date.
            updated_at_lte: Returned projects should have a label whose update date is lower or equal to this date.
            archived: If `True`, only archived projects are returned, if `False`, only active projects are returned.
                `None` disables this filter.
            starred: If `True`, only starred projects are returned, if `False`, only unstarred projects are returned.
                `None` disables this filter.
            tags_in: Returned projects should have at least one of these tags.
            organization_id: Returned projects should belong to this organization.
            fields: All the fields to request among the possible fields for the projects.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#project) for all possible fields.
            first: Maximum number of projects to return.
            skip: Number of projects to skip (they are ordered by their creation).
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the projects is returned.
            deleted: If `True`, all projects are returned (including deleted ones).

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            A list of projects or a generator of projects if `as_generator` is `True`.

        Examples:
            >>> # List all my projects
            >>> kili.projects()
        """
        tag_ids = (
            TagUseCases(self.kili_api_gateway).get_tag_ids_from_labels(tags_in) if tags_in else None
        )

        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)

        projects_gen = ProjectUseCases(self.kili_api_gateway).list_projects(
            ProjectFilters(
                id=ProjectId(project_id) if project_id else None,
                archived=archived,
                search_query=search_query,
                should_relaunch_kpi_computation=should_relaunch_kpi_computation,
                starred=starred,
                updated_at_gte=updated_at_gte,
                updated_at_lte=updated_at_lte,
                created_at_gte=None,
                created_at_lte=None,
                organization_id=organization_id,
                tag_ids=tag_ids,
                deleted=deleted,
            ),
            fields,
            options=QueryOptions(disable_tqdm=disable_tqdm, first=first, skip=skip),
        )

        if as_generator:
            return projects_gen
        return list(projects_gen)

    @typechecked
    # pylint: disable=too-many-arguments,too-many-locals
    def update_properties_in_project(
        self,
        project_id: str,
        can_navigate_between_assets: Optional[bool] = None,
        can_skip_asset: Optional[bool] = None,
        compliance_tags: Optional[ListOrTuple[ComplianceTag]] = None,
        consensus_mark: Optional[float] = None,
        consensus_tot_coverage: Optional[int] = None,
        description: Optional[str] = None,
        honeypot_mark: Optional[float] = None,
        instructions: Optional[str] = None,
        input_type: Optional[InputType] = None,
        json_interface: Optional[dict] = None,
        min_consensus_size: Optional[int] = None,
        number_of_assets: Optional[int] = None,
        number_of_skipped_assets: Optional[int] = None,
        number_of_remaining_assets: Optional[int] = None,
        number_of_reviewed_assets: Optional[int] = None,
        review_coverage: Optional[int] = None,
        should_relaunch_kpi_computation: Optional[bool] = None,
        title: Optional[str] = None,
        use_honeypot: Optional[bool] = None,
        metadata_types: Optional[dict] = None,
        seconds_to_label_before_auto_assign: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Update properties of a project.

        Args:
            project_id: Identifier of the project.
            can_navigate_between_assets:
                Activate / Deactivate the use of next and previous buttons in labeling interface.
            can_skip_asset: Activate / Deactivate the use of skip button in labeling interface.
            compliance_tags: Compliance tags of the project.
                Compliance tags are used to categorize projects based on the sensitivity of
                the data being handled and the legal constraints associated with it.
                Possible values are: `PHI` and `PII`.
            consensus_mark: Should be between 0 and 1.
            consensus_tot_coverage: Should be between 0 and 100.
                It is the percentage of the dataset that will be annotated several times.
            description: Description of the project.
            honeypot_mark: Should be between 0 and 1
            instructions: Instructions of the project.
            input_type: Currently, one of `IMAGE`, `PDF`, `TEXT` or `VIDEO`.
            json_interface: The json parameters of the project, see Edit your interface.
            min_consensus_size: Should be between 1 and 10
                Number of people that will annotate the same asset, for consensus computation.
            number_of_assets: Defaults to 0
            number_of_skipped_assets: Defaults to 0
            number_of_remaining_assets: Defaults to 0
            number_of_reviewed_assets: Defaults to 0
            review_coverage: Allow to set the percentage of assets
                that will be queued in the review interface.
                Should be between 0 and 100
            should_relaunch_kpi_computation: Technical field, added to indicate changes
                in honeypot or consensus settings
            title: Title of the project
            use_honeypot: Activate / Deactivate the use of honeypot in the project
            metadata_types: Types of the project metadata.
                Should be a `dict` of metadata fields name as keys and metadata types as values.
                Currently, possible types are: `string`, `number`
            seconds_to_label_before_auto_assign: When a user begins to annotate an asset,
                it automatically gets assigned to them.
                This ensures the user retains the asset until it is submitted.

        Returns:
            A dict with the changed properties which indicates if the mutation was successful,
                else an error message.

        !!! example "Change Metadata Types"
            Metadata fields are by default interpreted as `string` types. To change the type
            of a metadata field, you can use the `update_properties_in_project` function with the
            metadata_types argument. `metadata_types` is given as a dict of metadata field names
            as keys and metadata types as values.

            ```python
            kili.update_properties_in_project(
                project_id = project_id,
                metadata_types = {
                    'customConsensus': 'number',
                    'sensitiveData': 'string',
                    'uploadedFromCloud': 'string',
                    'modelLabelErrorScore': 'number'
                }
            )
            ```

            Not providing a type for a metadata field or providing an unsupported one
            will default to the `string` type.
        """
        return ProjectUseCases(self.kili_api_gateway).update_properties_in_project(
            ProjectId(project_id),
            can_navigate_between_assets=can_navigate_between_assets,
            can_skip_asset=can_skip_asset,
            compliance_tags=compliance_tags,
            consensus_mark=consensus_mark,
            consensus_tot_coverage=consensus_tot_coverage,
            description=description,
            honeypot_mark=honeypot_mark,
            instructions=instructions,
            input_type=input_type,
            json_interface=json_interface,
            min_consensus_size=min_consensus_size,
            number_of_assets=number_of_assets,
            number_of_skipped_assets=number_of_skipped_assets,
            number_of_remaining_assets=number_of_remaining_assets,
            number_of_reviewed_assets=number_of_reviewed_assets,
            review_coverage=review_coverage,
            should_relaunch_kpi_computation=should_relaunch_kpi_computation,
            use_honeypot=use_honeypot,
            title=title,
            metadata_types=metadata_types,
            seconds_to_label_before_auto_assign=seconds_to_label_before_auto_assign,
        )

    @typechecked
    # pylint: disable=too-many-arguments
    def count_projects(
        self,
        project_id: Optional[str] = None,
        search_query: Optional[str] = None,
        should_relaunch_kpi_computation: Optional[bool] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        archived: Optional[bool] = None,
        deleted: Optional[bool] = None,
    ) -> int:
        # pylint: disable=line-too-long
        """Count the number of projects with a search_query.

        Args:
            project_id: Select a specific project through its project_id.
            search_query: Returned projects with a title or a description matching this [PostgreSQL ILIKE](https://www.postgresql.org/docs/current/functions-matching.html#FUNCTIONS-LIKE) pattern.
            should_relaunch_kpi_computation: Technical field, added to indicate changes in honeypot
                or consensus settings
            updated_at_gte: Returned projects should have a label
                whose update date is greater
                or equal to this date.
            updated_at_lte: Returned projects should have a label
                whose update date is lower or equal to this date.
            archived: If `True`, only archived projects are returned, if `False`, only active projects are returned.
                None disable this filter.
            deleted: If `True` all projects are counted (including deleted ones).

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            The number of projects with the parameters provided
        """
        return ProjectUseCases(self.kili_api_gateway).count_projects(
            ProjectFilters(
                id=ProjectId(project_id) if project_id else None,
                search_query=search_query,
                should_relaunch_kpi_computation=should_relaunch_kpi_computation,
                updated_at_gte=updated_at_gte,
                updated_at_lte=updated_at_lte,
                archived=archived,
                deleted=deleted,
            )
        )



"""Kili Python SDK client."""

import getpass
import logging
import os
import sys
import warnings
from typing import Dict, Optional, Union

from kili.adapters.authentification import is_api_key_valid
from kili.adapters.http_client import HttpClient
from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.core.graphql.graphql_client import GraphQLClient, GraphQLClientName
from kili.entrypoints.mutations.asset import MutationsAsset
from kili.entrypoints.mutations.issue import MutationsIssue
from kili.entrypoints.mutations.notification import MutationsNotification
from kili.entrypoints.mutations.plugins import MutationsPlugins
from kili.entrypoints.mutations.project import MutationsProject
from kili.entrypoints.mutations.project_version import MutationsProjectVersion
from kili.entrypoints.queries.plugins import QueriesPlugins
from kili.entrypoints.queries.project_user import QueriesProjectUser
from kili.entrypoints.queries.project_version import QueriesProjectVersion
from kili.entrypoints.subscriptions.label import SubscriptionsLabel
from kili.exceptions import AuthenticationFailed
from kili.llm.presentation.client.llm import LlmClientMethods
from kili.presentation.client.asset import AssetClientMethods
from kili.presentation.client.cloud_storage import CloudStorageClientMethods
from kili.presentation.client.internal import InternalClientMethods
from kili.presentation.client.issue import IssueClientMethods
from kili.presentation.client.label import LabelClientMethods
from kili.presentation.client.notification import NotificationClientMethods
from kili.presentation.client.organization import OrganizationClientMethods
from kili.presentation.client.project import ProjectClientMethods
from kili.presentation.client.tag import TagClientMethods
from kili.presentation.client.user import UserClientMethods
from kili.use_cases.api_key import ApiKeyUseCases

warnings.filterwarnings("default", module="kili", category=DeprecationWarning)


class FilterPoolFullWarning(logging.Filter):
    """Filter out the specific urllib3 warning related to the connection pool."""

    def filter(self, record) -> bool:
        """urllib3.connectionpool:Connection pool is full, discarding connection: ..."""
        return "Connection pool is full, discarding connection" not in record.getMessage()


logging.getLogger("urllib3.connectionpool").addFilter(FilterPoolFullWarning())


class Kili(  # pylint: disable=too-many-ancestors,too-many-instance-attributes
    MutationsAsset,
    MutationsIssue,
    MutationsNotification,
    MutationsPlugins,
    MutationsProject,
    MutationsProjectVersion,
    QueriesPlugins,
    QueriesProjectUser,
    QueriesProjectVersion,
    SubscriptionsLabel,
    AssetClientMethods,
    CloudStorageClientMethods,
    IssueClientMethods,
    LabelClientMethods,
    NotificationClientMethods,
    OrganizationClientMethods,
    ProjectClientMethods,
    TagClientMethods,
    UserClientMethods,
):
    """Kili Client."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        api_key: Optional[str] = None,
        api_endpoint: Optional[str] = None,
        verify: Optional[Union[bool, str]] = None,
        client_name: GraphQLClientName = GraphQLClientName.SDK,
        graphql_client_params: Optional[Dict[str, object]] = None,
    ) -> None:
        """Initialize Kili client.

        Args:
            api_key: User API key generated
                from https://cloud.kili-technology.com/label/my-account/api-key.
                Default to `KILI_API_KEY` environment variable.
                If not passed, requires the `KILI_API_KEY` environment variable to be set.
            api_endpoint: Recipient of the HTTP operation.
                Default to `KILI_API_ENDPOINT` environment variable.
                If not passed, default to Kili SaaS:
                'https://cloud.kili-technology.com/api/label/v2/graphql'
            verify: similar to `requests`' verify.
                Either a boolean, in which case it controls whether we verify
                the server's TLS certificate, or a string, in which case it must be a path
                to a CA bundle to use. Defaults to ``True``. When set to
                ``False``, requests will accept any TLS certificate presented by
                the server, and will ignore hostname mismatches and/or expired
                certificates, which will make your application vulnerable to
                man-in-the-middle (MitM) attacks. Setting verify to ``False``
                may be useful during local development or testing.
            client_name: For internal use only.
                Define the name of the graphQL client whith which graphQL calls will be sent.
            graphql_client_params: Parameters to pass to the graphQL client.

        Returns:
            Instance of the Kili client.

        Examples:
            ```python
            from kili.client import Kili

            kili = Kili()

            kili.assets()  # list your assets
            kili.labels()  # list your labels
            kili.projects()  # list your projects
            ```
        """
        api_key = api_key or os.getenv("KILI_API_KEY")

        if not api_key and sys.stdin.isatty():
            api_key = getpass.getpass(
                "No `KILI_API_KEY` environment variable found.\nPlease enter your API key: "
            )

        if api_endpoint is None:
            api_endpoint = os.getenv(
                "KILI_API_ENDPOINT",
                "https://cloud.kili-technology.com/api/label/v2/graphql",
            )

        if not api_key:
            raise AuthenticationFailed(api_key, api_endpoint)

        if verify is None:
            verify = os.getenv(
                "KILI_VERIFY",
                "True",
            ).lower() in ("true", "1", "yes")

        self.api_key = api_key
        self.api_endpoint = api_endpoint
        self.verify = verify
        self.client_name = client_name
        self.http_client = HttpClient(kili_endpoint=api_endpoint, verify=verify, api_key=api_key)
        skip_checks = os.getenv("KILI_SDK_SKIP_CHECKS") is not None
        if not skip_checks and not is_api_key_valid(
            self.http_client, api_key, api_endpoint, client_name
        ):
            raise AuthenticationFailed(
                api_key=self.api_key,
                api_endpoint=self.api_endpoint,
                error_msg="Api key does not seem to be valid.",
            )

        self.graphql_client = GraphQLClient(
            endpoint=api_endpoint,
            api_key=api_key,
            client_name=client_name,
            verify=self.verify,
            http_client=self.http_client,
            **(graphql_client_params or {}),  # pyright: ignore[reportGeneralTypeIssues]
        )
        self.kili_api_gateway = KiliAPIGateway(self.graphql_client, self.http_client)
        self.internal = InternalClientMethods(self.kili_api_gateway)
        self.llm = LlmClientMethods(self.kili_api_gateway)

        if not skip_checks:
            api_key_use_cases = ApiKeyUseCases(self.kili_api_gateway)
            api_key_use_cases.check_expiry_of_key_is_close(api_key)



"""GraphQL Asset operations."""


def get_api_keys_query(fragment: str) -> str:
    """Return the GraphQL apiKeys query."""
    return f"""
        query apiKeys($where: ApiKeyWhere!, $first: PageSize!, $skip: Int!) {{
          data: apiKeys(where: $where, skip: $skip, first: $first) {{
            {fragment}
          }}
        }}
        """


GQL_COUNT_API_KEYS = """
query countApiKeys($where: ApiKeyWhere!) {
    data: countApiKeys(where: $where)
}
"""

GQL_API_KEY_EXPIRY_DATE = """
    query apiKeys($where: ApiKeyWhere!) {
        data: apiKeys(where: $where, skip: 0, first: 1) {
            expiryDate
        }
    }
    """



"""Logger utils."""

import logging

from kili.services.types import LogLevel


def get_logger(level: LogLevel) -> logging.Logger:
    """Gets the export logger."""
    logger = logging.getLogger("kili.services.export")
    logger.setLevel(level)
    if logger.hasHandlers():
        logger.handlers.clear()
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    return logger



"""Constants for the asset_import services."""

from kili.core.constants import mime_extensions_for_IV2

project_compatible_mimetypes = mime_extensions_for_IV2

IMPORT_BATCH_SIZE = 100
FRAME_IMPORT_BATCH_SIZE = 1

MB_SIZE = 1024**2
LARGE_IMAGE_THRESHOLD_SIZE = 30 * MB_SIZE



"""Project member command of Kili CLI."""

import click

from kili.entrypoints.cli.common_args import CONTEXT_SETTINGS
from kili.entrypoints.cli.project.member.add import add_member
from kili.entrypoints.cli.project.member.list_ import list_members
from kili.entrypoints.cli.project.member.remove import remove_member
from kili.entrypoints.cli.project.member.update import update_member


@click.group(context_settings=CONTEXT_SETTINGS)
def member():
    """Commands to interact with Kili project members."""


member.add_command(list_members, name="list")
member.add_command(add_member, name="add")
member.add_command(update_member, name="update")
member.add_command(remove_member, name="rm")



"""Module for methods that are for internal use by Kili Technology only."""

from typing import Dict, Iterable, Optional

from typeguard import typechecked

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway
from kili.adapters.kili_api_gateway.user.operations import get_reset_password_mutation
from kili.domain.api_key import ApiKeyFilters
from kili.domain.types import ListOrTuple
from kili.entrypoints.mutations.project.queries import GQL_DELETE_PROJECT
from kili.presentation.client.organization import InternalOrganizationClientMethods
from kili.use_cases.api_key import ApiKeyUseCases


class InternalClientMethods(InternalOrganizationClientMethods):
    """Kili client methods for internal use by Kili Technology only."""

    def __init__(self, kili_api_gateway: KiliAPIGateway) -> None:
        """Initialize the class.

        Args:
            kili: Kili object
        """
        super().__init__()
        self.kili_api_gateway = kili_api_gateway

    @typechecked
    def reset_password(self, email: str):
        """Reset password.

        WARNING: This method is for internal use only.

        Args:
            email: Email of the person whose password has to be reset.

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        query = get_reset_password_mutation(fragment="id")
        variables = {"where": {"email": email}}
        result = self.kili_api_gateway.graphql_client.execute(query, variables)
        return result["data"]

    @typechecked
    def delete_project(self, project_id: str):
        """Delete project permanently.

        WARNING: This resolver is for internal use by Kili Technology only.

        Args:
            project_id: Identifier of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {"projectID": project_id}
        result = self.kili_api_gateway.graphql_client.execute(GQL_DELETE_PROJECT, variables)
        return result["data"]

    # pylint: disable=too-many-arguments
    @typechecked
    def api_keys(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
        skip: int = 0,
        fields: ListOrTuple[str] = ("id", "name", "createdAt", "revoked"),
        first: Optional[int] = None,
        disable_tqdm: Optional[bool] = None,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        """Get a generator or a list of API keys that match a set of constraints.

        WARNING: This method is for internal use only.

        !!! info
            You can only query your own API keys

        Args:
            api_key_id: Identifier of the API key to retrieve.
            user_id: Identifier of the user.
            api_key: Value of the API key.
            skip: Number of assets to skip
            fields: All the fields to request among the possible fields for the assets.
            first: Maximum number of API keys to return.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the API key is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.
        """
        api_key_use_cases = ApiKeyUseCases(self.kili_api_gateway)
        filters = ApiKeyFilters(api_key_id=api_key_id, user_id=user_id, api_key=api_key)
        api_keys_gen = api_key_use_cases.list_api_keys(
            filters, fields, QueryOptions(first=first, skip=skip, disable_tqdm=disable_tqdm)
        )

        if as_generator:
            return api_keys_gen
        return list(api_keys_gen)

    @typechecked
    def count_api_keys(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
    ) -> int:
        """Count and return the number of api keys with the given constraints.

        WARNING: This method is for internal use only.

        Args:
            api_key_id: Identifier of the API key to retrieve.
            user_id: Identifier of the user.
            api_key: Value of the api key.

        Returns:
            The number of API Keys matching params if it was successful,
                or an error message.
        """
        api_key_use_cases = ApiKeyUseCases(self.kili_api_gateway)
        filters = ApiKeyFilters(api_key_id=api_key_id, user_id=user_id, api_key=api_key)
        return api_key_use_cases.count_api_keys(filters)



"""Utils for use cases."""

from itertools import chain
from typing import TYPE_CHECKING, Dict, Optional

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.core.utils import pagination
from kili.domain.asset import AssetExternalId, AssetFilters, AssetId
from kili.domain.asset.helpers import check_asset_identifier_arguments
from kili.domain.project import ProjectId
from kili.domain.types import ListOrTuple
from kili.exceptions import NotFound

if TYPE_CHECKING:
    from kili.adapters.kili_api_gateway.kili_api_gateway import KiliAPIGateway


class AssetUseCasesUtils:
    """Utils for use cases."""

    def __init__(self, kili_api_gateway: "KiliAPIGateway") -> None:
        """Init."""
        self.kili_api_gateway = kili_api_gateway

    def get_asset_ids_or_throw_error(
        self,
        asset_ids: Optional[ListOrTuple[AssetId]],
        external_ids: Optional[ListOrTuple[AssetExternalId]],
        project_id: Optional[ProjectId],
    ) -> ListOrTuple[AssetId]:
        """Check if external id to internal id conversion is valid and needed."""
        check_asset_identifier_arguments(project_id, asset_ids, external_ids)

        if asset_ids is None and external_ids is not None and project_id is not None:
            id_map = self.infer_ids_from_external_ids(external_ids, project_id)
            resolved_asset_ids = [id_map[ext_id] for ext_id in external_ids]
        else:
            resolved_asset_ids = asset_ids or []

        return resolved_asset_ids

    def infer_ids_from_external_ids(
        self, asset_external_ids: ListOrTuple[AssetExternalId], project_id: ProjectId
    ) -> Dict[AssetExternalId, AssetId]:
        """Infer asset ids from their external ids and project Id.

        Args:
            asset_external_ids: asset external ids
            project_id: project id

        Returns:
            a dict that maps external ids to internal ids.

        Raises:
            NotFound: when there are asset_ids what have the same external id, or when external ids
            have not been found.
        """
        id_map = self._build_id_map(asset_external_ids, project_id)

        if len(id_map) < len(set(asset_external_ids)):
            assets_not_found = [
                external_id for external_id in asset_external_ids if external_id not in id_map
            ]
            raise NotFound(
                f"The assets whose external_id are: {assets_not_found} have not been found in the"
                f" project of Id {project_id}"
            )
        if len(id_map) > len(set(asset_external_ids)):
            raise NotFound(
                "Several assets have been found for the same external_id. Please consider using"
                " asset ids instead."
            )
        return id_map

    def _build_id_map(
        self, asset_external_ids: ListOrTuple[AssetExternalId], project_id: ProjectId
    ) -> Dict[AssetExternalId, AssetId]:
        # we batch the queries because too many assets in a "in" query makes the query fail
        assets_generators = (
            self.kili_api_gateway.list_assets(
                AssetFilters(project_id, external_id_strictly_in=external_ids_batch),
                ["id", "externalId"],
                QueryOptions(disable_tqdm=True),
            )
            for external_ids_batch in pagination.batcher(list(asset_external_ids), 1000)
        )
        assets = chain(*assets_generators)
        id_map: Dict[AssetExternalId, AssetId] = {}
        asset_external_ids_set = set(asset_external_ids)
        for asset in (asset for asset in assets if asset["externalId"] in asset_external_ids_set):
            id_map[AssetExternalId(asset["externalId"])] = AssetId(asset["id"])
        return id_map



"""Api Keys use cases."""

import warnings
from datetime import datetime, timedelta
from typing import Dict, Generator

from kili.adapters.kili_api_gateway.helpers.queries import QueryOptions
from kili.domain.api_key import ApiKeyFilters
from kili.domain.types import ListOrTuple
from kili.use_cases.base import BaseUseCases


class ApiKeyUseCases(BaseUseCases):
    """Api Keys use cases."""

    def list_api_keys(
        self, filters: ApiKeyFilters, fields: ListOrTuple[str], options: QueryOptions
    ) -> Generator[Dict, None, None]:
        """List api keys with given options."""
        return self._kili_api_gateway.list_api_keys(filters, fields, options)

    def count_api_keys(self, filters: ApiKeyFilters) -> int:
        """Count api keys with given options."""
        return self._kili_api_gateway.count_api_keys(filters)

    def check_expiry_of_key_is_close(self, api_key: str) -> None:
        """Check that the expiration date of the api_key is not too close."""
        warn_days = 30

        key_expiry = self._kili_api_gateway.get_api_key_expiry_date(api_key)
        key_remaining_time = key_expiry - datetime.now()
        key_soon_deprecated = key_remaining_time < timedelta(days=warn_days)
        if key_soon_deprecated:
            message = f"""
                Your api key will be deprecated on {key_expiry:%Y-%m-%d}.
                You should generate a new one on My account > API KEY."""
            warnings.warn(message, UserWarning, stacklevel=2)



"""WebsocketGraphQL Client."""

import json
import random
import string
import threading
import time
from datetime import datetime
from typing import Callable, Dict, Optional

import websocket
from typing_extensions import LiteralString

GQL_WS_SUBPROTOCOL = "graphql-ws"


class SubscriptionGraphQLClient:
    """A simple GraphQL client that works over Websocket as the transport protocol, instead of HTTP.

    This follows the Apollo protocol.
    https://github.com/apollographql/subscriptions-transport-ws/blob/master/PROTOCOL.md
    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments

    def __init__(self, url) -> None:
        self.ws_url = url
        self._id = None
        self._paused = False
        self._connect()
        self._subscription_running = False
        self._st_id = None
        self.failed_connection_attempts = 0

    def _connect(self):
        """Handles the connection."""
        self._conn = websocket.create_connection(
            self.ws_url, on_message=self._on_message, subprotocols=[GQL_WS_SUBPROTOCOL]
        )
        self._created_at = datetime.now()
        self._conn.on_message = self._on_message  # type: ignore

    def _reconnect(self):
        """Handles the reconnection."""
        self._connect()
        self._subscription_running = True
        dt_string = datetime.now().strftime(r"%d/%m/%Y %H:%M:%S")
        print(f"{dt_string} reconnected")
        self.failed_connection_attempts = 0

    def _on_message(self, message):
        """Handles messages.

        Args:
            message : the message
        """
        data = json.loads(message)
        # skip keepalive messages
        if data["type"] != "ka":
            print(message)

    def _conn_init(self, headers=None, authorization=None):
        """Initializes the websocket connection.

        Args:
            headers : Headers are necessary for Kili API v1
            authorization : Headers are necessary for Kili API v2
        """
        payload = {
            "type": "connection_init",
            "payload": {"headers": headers, "Authorization": authorization},
        }
        self._conn.send(json.dumps(payload).encode("utf-8"))
        self._conn.recv()

    def _start(self, payload):
        """Handles start.

        Args:
            payload
        """
        _id = gen_id()
        frame = {"id": _id, "type": "start", "payload": payload}
        self._conn.send(json.dumps(frame).encode("utf-8"))
        return _id

    def _stop(self, _id):
        """Handles stop.

        Args:
        - _id: connection id
        """
        payload = {"id": _id, "type": "stop"}
        self._conn.send(json.dumps(payload).encode("utf-8"))
        return self._conn.recv()

    def query(self, query: str, variables: Optional[Dict] = None, headers: Optional[Dict] = None):
        """Sends a query.

        Args:
            query: the GraphQL query
            variables: the payload of the query
            headers: headers
        """
        self._conn_init(headers)
        payload = {"headers": headers, "query": query, "variables": variables}
        _id = self._start(payload)
        res = self._conn.recv()
        self._stop(_id)
        return res

    def prepare_subscribe(
        self,
        query: str,
        variables: Optional[Dict],
        headers: Optional[Dict],
        callback: Optional[Callable],
        authorization: Optional[str],
    ):
        """Prepares a subscription.

        Args:
            query: the GraphQL query
            variables: the payload of the query
            headers: headers
            callback: function executed after the subscription
            authorization: authorization header
        """
        self._conn_init(headers, authorization)
        payload = {"headers": headers, "query": query, "variables": variables}
        _cc = callback if callback else self._on_message
        _id = self._start(payload)
        self._id = _id
        return _cc, _id

    def subscribe(
        self,
        query: str,
        variables: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        callback: Optional[Callable] = None,
        authorization: Optional[str] = None,
    ):
        """Subscribes.

        Args:
            query: the GraphQL query
            variables: the payload of the query
            headers: headers
            callback: function executed after the subscription
            authorization: authorization header
        """
        _cc, _id = self.prepare_subscribe(query, variables, headers, callback, authorization)

        def subs(_cc, _id):
            max_reconnections = 10
            self._subscription_running = True
            while (
                self._subscription_running and self.failed_connection_attempts < max_reconnections
            ):
                try:
                    response = json.loads(self._conn.recv())
                    if response["type"] == "error" or response["type"] == "complete":
                        print(response)
                        self._stop_subscribe(_id)
                        break
                    if response["type"] != "ka" and not self._paused:
                        _cc(_id, response)  # type:ignore
                    time.sleep(1)
                except (
                    websocket._exceptions.WebSocketConnectionClosedException  # pylint: disable=protected-access
                ) as error:
                    self.failed_connection_attempts += 1
                    dt_string = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                    error_message = str(error)
                    print(f"{dt_string} Connection closed error : {error_message}")
                    print(
                        "Will try to reconnect"
                        f" {max_reconnections - self.failed_connection_attempts} times..."
                    )
                    self._reconnect()
                    _cc, _id = self.prepare_subscribe(
                        query, variables, headers, callback, authorization
                    )
                    continue
            print(f"Did not reconnect successfully after {max_reconnections} attempts")

        self._st_id = threading.Thread(target=subs, args=(_cc, _id))
        self._st_id.start()
        return _id

    def _stop_subscribe(self, _id):
        self._subscription_running = False
        self._stop(_id)

    def close(self):
        """Handles close."""
        self._conn.close()

    def pause(self):
        """Handles pause."""
        self._paused = True

    def unpause(self):
        """Handles unpause."""
        self._paused = False

    def get_lifetime(self):
        """Return the lifetime."""
        return (datetime.now() - self._created_at).seconds

    def reset_timeout(self):
        """Resets the timeout."""
        self._reconnect()


def gen_id(size: int = 6, chars: LiteralString = string.ascii_letters + string.digits) -> str:
    """Generate random alphanumeric id.

    Args:
        size: length of the id
        chars: chars used to generate the id
    """
    return "".join(random.choice(chars) for _ in range(size))



"""Mixin extending Kili API Gateway class with notification-related operations."""

from typing import Dict, Generator

from kili.adapters.kili_api_gateway.base import BaseOperationMixin
from kili.adapters.kili_api_gateway.helpers.queries import (
    PaginatedGraphQLQuery,
    QueryOptions,
    fragment_builder,
)
from kili.domain.notification import NotificationFilter
from kili.domain.types import ListOrTuple

from .mappers import map_notification_filter
from .operations import GQL_COUNT_NOTIFICATIONS, get_notifications_query


class NotificationOperationMixin(BaseOperationMixin):
    """GraphQL Mixin extending GraphQL Gateway class with notification-related operations."""

    def list_notifications(
        self, filters: NotificationFilter, fields: ListOrTuple[str], options: QueryOptions
    ) -> Generator[Dict, None, None]:
        """List notifications."""
        fragment = fragment_builder(fields)
        query = get_notifications_query(fragment)
        where = map_notification_filter(filters=filters)
        return PaginatedGraphQLQuery(self.graphql_client).execute_query_from_paginated_call(
            query, where, options, "Retrieving notifications", GQL_COUNT_NOTIFICATIONS
        )

    def count_notification(self, filters: NotificationFilter) -> int:
        """Count notifications."""
        variables = {"where": map_notification_filter(filters=filters)}
        return self.graphql_client.execute(GQL_COUNT_NOTIFICATIONS, variables)["data"]
